# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
Providing mechanisms to C{X2GoControlSession*} backends for checking host validity.

"""
__NAME__ = 'x2gocheckhosts-pylib'

# modules
import paramiko
import binascii

# Python X2Go modules
import sshproxy
import log
import x2go_exceptions
import random
import string

class X2GoMissingHostKeyPolicy(paramiko.MissingHostKeyPolicy):
    """\
    Skeleton class for Python X2Go's missing host key policies.

    """
    def __init__(self, caller=None, session_instance=None, fake_hostname=None):
        """\
        @param caller: calling instance
        @type caller: C{class}
        @param session_instance: an X2Go session instance
        @type session_instance: L{X2GoSession} instance

        """
        self.caller = caller
        self.session_instance = session_instance
        self.fake_hostname = fake_hostname

    def get_client(self):
        """\
        Retrieve the Paramiko SSH/Client.

        @return: the associated X2Go control session instance.
        @rtype: C{X2GoControlSession*} instance

        """
        return self.client

    def get_hostname(self):
        """\
        Retrieve the server hostname:port expression of the server to be validated.

        @return: hostname:port
        @rtype: C{str}

        """
        return self.fake_hostname or self.hostname

    def get_hostname_name(self):
        """\
        Retrieve the server hostname string of the server to be validated.

        @return: hostname
        @rtype: C{str}

        """
        return self.get_hostname().split(':')[0].lstrip('[').rstrip(']')

    def get_hostname_port(self):
        """\
        Retrieve the server port of the server to be validated.

        @return: port
        @rtype: C{str}

        """
        return self.get_hostname().split(':')[1]

    def get_key(self):
        """\
        Retrieve the host key of the server to be validated.

        @return: host key
        @rtype: Paramiko/SSH key instance

        """
        return self.key

    def get_key_name(self):
        """\
        Retrieve the host key name of the server to be validated.

        @return: host key name (RSA, DSA, ...)
        @rtype: C{str}

        """
        return self.key.get_name().upper()

    def get_key_fingerprint(self):
        """\
        Retrieve the host key fingerprint of the server to be validated.

        @return: host key fingerprint
        @rtype: C{str}

        """
        return binascii.hexlify(self.key.get_fingerprint())

    def get_key_fingerprint_with_colons(self):
        """\
        Retrieve the (colonized) host key fingerprint of the server
        to be validated.

        @return: host key fingerprint (with colons)
        @rtype: C{str}

        """
        _fingerprint = self.get_key_fingerprint()
        _colon_fingerprint = ''
        idx = 0
        for char in _fingerprint:
            idx += 1
            _colon_fingerprint += char
            if idx % 2 == 0:
                _colon_fingerprint += ':'
        return _colon_fingerprint.rstrip(':')


class X2GoAutoAddPolicy(X2GoMissingHostKeyPolicy):

    def missing_host_key(self, client, hostname, key):
        self.client = client
        self.hostname = hostname
        self.key = key
        if self.session_instance and self.session_instance.control_session.unique_hostkey_aliases:
            self.client._host_keys.add(self.session_instance.get_profile_id(), self.key.get_name(), self.key)
        else:
            self.client._host_keys.add(self.get_hostname(), self.key.get_name(), self.key)
        if self.client._host_keys_filename is not None:
            self.client.save_host_keys(self.client._host_keys_filename)
        self.client._log(paramiko.common.DEBUG, 'Adding %s host key for %s: %s' %
                         (self.key.get_name(), self.get_hostname(), binascii.hexlify(self.key.get_fingerprint())))


class X2GoInteractiveAddPolicy(X2GoMissingHostKeyPolicy):
    """\
    Policy for making host key information available to Python X2Go after a
    Paramiko/SSH connect has been attempted. This class needs information
    about the associated L{X2GoSession} instance.

    Once called, the L{missing_host_key} method of this class will try to call
    L{X2GoSession.HOOK_check_host_dialog()}. This hook method---if not re-defined
    in your application---will then try to call the L{X2GoClient.HOOK_check_host_dialog()},
    which then will return C{True} by default if not customized in your application.

    To accept host key checks, make sure to either customize the 
    L{X2GoClient.HOOK_check_host_dialog()} method or the L{X2GoSession.HOOK_check_host_dialog()}
    method and hook some interactive user dialog to either of them.

    """
    def missing_host_key(self, client, hostname, key):
        """\
        Handle a missing host key situation. This method calls

        Once called, the L{missing_host_key} method will try to call
        L{X2GoSession.HOOK_check_host_dialog()}. This hook method---if not re-defined
        in your application---will then try to call the L{X2GoClient.HOOK_check_host_dialog()},
        which then will return C{True} by default if not customized in your application.

        To accept host key checks, make sure to either customize the 
        L{X2GoClient.HOOK_check_host_dialog()} method or the L{X2GoSession.HOOK_check_host_dialog()}
        method and hook some interactive user dialog to either of them.

        @param client: SSH client (C{X2GoControlSession*}) instance
        @type client: C{X2GoControlSession*} instance
        @param hostname: remote hostname
        @type hostname: C{str}
        @param key: host key to validate
        @type key: Paramiko/SSH key instance

        @raise X2GoHostKeyException: if the X2Go server host key is not in the C{known_hosts} file
        @raise X2GoSSHProxyHostKeyException: if the SSH proxy host key is not in the C{known_hosts} file
        @raise SSHException: if this instance does not know its {self.session_instance}

        """
        self.client = client
        self.hostname = hostname
        if (self.hostname.find(']') == -1) and (self.hostname.find(':') == -1):
            # if hostname is an IPv4 quadruple with standard SSH port...
            self.hostname = '[%s]:22' % self.hostname
        self.key = key
        self.client._log(paramiko.common.DEBUG, 'Interactively Checking %s host key for %s: %s' %
                         (self.key.get_name(), self.get_hostname(), binascii.hexlify(self.key.get_fingerprint())))
        if self.session_instance:

            if self.fake_hostname is not None:
                server_key = client.get_transport().get_remote_server_key()
                keytype = server_key.get_name()
                our_server_key = client._system_host_keys.get(self.fake_hostname, {}).get(keytype, None)
                if our_server_key is None:
                    if self.session_instance.control_session.unique_hostkey_aliases:
                        our_server_key = client._host_keys.get(self.session_instance.get_profile_id(), {}).get(keytype, None)
                        if our_server_key is not None:
                            self.session_instance.logger('SSH host key verification for SSH-proxied host %s with %s fingerprint ,,%s\'\' succeeded. This host is known by the X2Go session profile ID of profile »%s«.' % (self.fake_hostname, self.get_key_name(), self.get_key_fingerprint_with_colons(), self.session_instance.profile_name), loglevel=log.loglevel_NOTICE)
                            return
                    else:
                        our_server_key = client._host_keys.get(self.fake_hostname, {}).get(keytype, None)
                        if our_server_key is not None:
                            self.session_instance.logger('SSH host key verification for SSH-proxied host %s with %s fingerprint ,,%s\'\' succeeded. This host is known by the address it has behind the SSH proxy host.' % (self.fake_hostname, self.get_key_name(), self.get_key_fingerprint_with_colons()), loglevel=log.loglevel_NOTICE)
                            return

            self.session_instance.logger('SSH host key verification for host %s with %s fingerprint ,,%s\'\' initiated. We are seeing this X2Go server for the first time.' % (self.get_hostname(), self.get_key_name(), self.get_key_fingerprint_with_colons()), loglevel=log.loglevel_NOTICE)
            _valid = self.session_instance.HOOK_check_host_dialog(self.get_hostname_name(),
                                                                  port=self.get_hostname_port(),
                                                                  fingerprint=self.get_key_fingerprint_with_colons(),
                                                                  fingerprint_type=self.get_key_name(),
                                                                 )
            if _valid:
                if self.session_instance.control_session.unique_hostkey_aliases and type(self.caller) not in (sshproxy.X2GoSSHProxy, ):
                    paramiko.AutoAddPolicy().missing_host_key(client, self.session_instance.get_profile_id(), key)
                else:
                    paramiko.AutoAddPolicy().missing_host_key(client, self.get_hostname(), key)

            else:
                if type(self.caller) in (sshproxy.X2GoSSHProxy, ):
                    raise x2go_exceptions.X2GoSSHProxyHostKeyException('Invalid host %s is not authorized for access. Add the host to Paramiko/SSH\'s known_hosts file.' % self.get_hostname())
                else:
                    raise x2go_exceptions.X2GoHostKeyException('Invalid host %s is not authorized for access. Add the host to Paramiko/SSH\'s known_hosts file.' % self.get_hostname())
        else:
            raise x2go_exceptions.SSHException('Policy has collected host key information on %s for further introspection' % self.get_hostname())


def check_ssh_host_key(x2go_sshclient_instance, hostname, port=22):
    """\
    Perform a Paramiko/SSH host key check by connecting to the host and
    validating the results (i.e. by validating raised exceptions during the
    connect process).

    @param x2go_sshclient_instance: a Paramiko/SSH client instance to be used for testing host key validity.
    @type x2go_sshclient_instance: C{X2GoControlSession*} instance
    @param hostname: hostname of server to validate
    @type hostname: C{str}
    @param port: port of server to validate
    @type port: C{int}

    @return: returns a tuple with the following components (<host_ok>, <hostname>, <port>, <fingerprint>, <fingerprint_type>)
    @rtype: C{tuple}

    @raise SSHException: if an SSH exception occurred, that we did not provocate in L{X2GoInteractiveAddPolicy.missing_host_key()}

    """
    _hostname = hostname
    _port = port
    _fingerprint = 'NO-FINGERPRINT'
    _fingerprint_type = 'SOME-KEY-TYPE'

    _check_policy = X2GoInteractiveAddPolicy()
    x2go_sshclient_instance.set_missing_host_key_policy(_check_policy)

    host_ok = False
    try:
        paramiko.SSHClient.connect(x2go_sshclient_instance, hostname=hostname, port=port, username='foo', password="".join([random.choice(string.letters+string.digits) for x in range(1, 20)]))
    except x2go_exceptions.AuthenticationException:
        host_ok = True
        x2go_sshclient_instance.logger('SSH host key verification for host [%s]:%s succeeded. Host is already known to the client\'s Paramiko/SSH sub-system.' % (_hostname, _port), loglevel=log.loglevel_NOTICE)
    except x2go_exceptions.SSHException, e:
        msg = str(e)
        if msg.startswith('Policy has collected host key information on '):
            _hostname = _check_policy.get_hostname().split(':')[0].lstrip('[').rstrip(']')
            _port = _check_policy.get_hostname().split(':')[1]
            _fingerprint = _check_policy.get_key_fingerprint_with_colons()
            _fingerprint_type = _check_policy.get_key_name()
        else:
            raise(e)
        x2go_sshclient_instance.set_missing_host_key_policy(paramiko.RejectPolicy())
    except:
        # let any other error be handled by subsequent algorithms
        pass

    return (host_ok, _hostname, _port, _fingerprint, _fingerprint_type)



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoProxyBASE class - proxying your connection through NX3 and others.

"""
__NAME__ = 'x2goproxy-pylib'

# modules
import gevent
import os
import copy
import threading
import socket

# Python X2Go modules
import x2go.forward as forward
import x2go.log as log
import x2go.utils as utils
import x2go.x2go_exceptions as x2go_exceptions

from x2go.defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS
if _X2GOCLIENT_OS in ("Windows"):
    import subprocess
else:
    import x2go.gevent_subprocess as subprocess
    from x2go.x2go_exceptions import WindowsError

from x2go.defaults import LOCAL_HOME as _LOCAL_HOME
from x2go.defaults import X2GO_SESSIONS_ROOTDIR as _X2GO_SESSIONS_ROOTDIR


class X2GoProxyBASE(threading.Thread):
    """\
    X2GoProxy is an abstract class for X2Go proxy connections.

    This class needs to be inherited from a concrete proxy class. Only 
    currently available proxy class is: L{X2GoProxyNX3}.

    """
    PROXY_CMD = ''
    """Proxy command. Needs to be set by a potential child class, might be OS specific."""
    PROXY_ARGS = []
    """Arguments to be passed to the proxy command. This needs to be set by a potential child class."""
    PROXY_ENV = {}
    """Provide environment variables to the proxy command. This also needs to be set by a child class."""

    session_info = None
    session_log_stdout = None
    session_log_stderr = None
    fw_tunnel = None
    proxy = None

    def __init__(self, session_info=None, 
                 ssh_transport=None, session_log="session.log", 
                 sessions_rootdir=os.path.join(_LOCAL_HOME, _X2GO_SESSIONS_ROOTDIR),
                 proxy_options={},
                 session_instance=None,
                 logger=None, loglevel=log.loglevel_DEFAULT, ):
        """\
        @param session_info: session information provided as an C{X2GoServerSessionInfo*} backend
            instance
        @type session_info: C{X2GoServerSessionInfo*} instance
        @param ssh_transport: SSH transport object from C{paramiko.SSHClient}
        @type ssh_transport: C{paramiko.Transport} instance
        @param session_log: name of the proxy's session logfile
        @type session_log: C{str}
        @param sessions_rootdir: base dir where X2Go session files are stored (by default: ~/.x2go)
        @type sessions_rootdir: C{str}
        @param proxy_options: a set of very C{X2GoProxy*} backend specific options; any option that is not known
            to the C{X2GoProxy*} backend will simply be ignored
        @type proxy_options: C{dict}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoProxy} constructor
        @param session_instance: the L{X2GoSession} instance this C{X2GoProxy*} instance belongs to
        @type session_instance: L{X2GoSession} instance
        @type logger: L{X2GoLogger} instance
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: int

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.sessions_rootdir = sessions_rootdir
        self.session_info = session_info
        self.session_name = self.session_info.name
        self.ssh_transport = ssh_transport
        self.session_log = session_log
        self.proxy_options = proxy_options
        self.session_instance = session_instance
        self.PROXY_ENV = os.environ.copy()
        self.proxy = None

        threading.Thread.__init__(self)
        self.daemon = True

    def __del__(self):
        """\
        On instance destruction make sure this proxy thread is stopped properly.

        """
        self.stop_thread()

    def _tidy_up(self):
        """\
        Close any left open port forwarding tunnel, also close session log file,
        if left open.

        """
        if self.proxy:
            self.logger('Shutting down X2Go proxy subprocess', loglevel=log.loglevel_DEBUG)
            try:
                self.proxy.kill()
            except OSError, e:
                self.logger('X2Go proxy shutdown gave a message that we may ignore: %s' % str(e), loglevel=log.loglevel_WARN)
            self.proxy = None
        if self.fw_tunnel is not None:
            self.logger('Shutting down Paramiko/SSH forwarding tunnel', loglevel=log.loglevel_DEBUG)
            forward.stop_forward_tunnel(self.fw_tunnel)
            self.fw_tunnel = None
        if self.session_log_stdout is not None:
            self.session_log_stdout.close()
        if self.session_log_stderr is not None:
            self.session_log_stderr.close()

    def stop_thread(self):
        """\
        End the thread runner and tidy up.

        """
        self._keepalive = False
        # wait for thread loop to finish...
        gevent.sleep(.5)
        self._tidy_up()

    def run(self):
        """\
        Start the X2Go proxy command. The X2Go proxy command utilizes a
        Paramiko/SSH based forwarding tunnel (openssh -L option). This tunnel
        gets started here and is forked into background (Greenlet/gevent).

        """
        self._keepalive = True
        self.proxy = None

        if self.session_info is None or self.ssh_transport is None:
            return None

        try:
            os.makedirs(self.session_info.local_container)
        except OSError, e:
            if e.errno == 17:
                # file exists
                pass

        local_graphics_port = self.session_info.graphics_port
        try:
            if self.ssh_transport.getpeername() in ('::1', '127.0.0.1', 'localhost', 'localhost.localdomain'):
                local_graphics_port += 10000
        except socket.error:
            raise x2go_exceptions.X2GoControlSessionException('The control session has died unexpectedly.')
        local_graphics_port = utils.detect_unused_port(preferred_port=local_graphics_port)

        self.fw_tunnel = forward.start_forward_tunnel(local_port=local_graphics_port,
                                                      remote_port=self.session_info.graphics_port,
                                                      ssh_transport=self.ssh_transport,
                                                      session_instance=self.session_instance,
                                                      session_name=self.session_name,
                                                      logger=self.logger,
                                                     )

        # update the proxy port in PROXY_ARGS
        self._update_local_proxy_socket(local_graphics_port)
        cmd_line = self._generate_cmdline()

        self.session_log_stdout = open('%s/%s' % (self.session_info.local_container, self.session_log, ), 'a')
        self.session_log_stderr = open('%s/%s' % (self.session_info.local_container, self.session_log, ), 'a')
        self.logger('forking threaded subprocess: %s' % " ".join(cmd_line), loglevel=log.loglevel_DEBUG)

        _stdin = None
        _shell = False
        if _X2GOCLIENT_OS == 'Windows':
            _stdin = file('nul', 'r')
            _shell = True

        # allow inheriting classes to do something with backend specific proxy_options...
        self.process_proxy_options()

        while not self.proxy:
            gevent.sleep(.2)
            p = self.proxy = subprocess.Popen(cmd_line,
                                              env=self.PROXY_ENV,
                                              stdin=_stdin,
                                              stdout=self.session_log_stdout,
                                              stderr=self.session_log_stderr,
                                              shell=_shell)

        while self._keepalive:
            gevent.sleep(.2)

        if _X2GOCLIENT_OS == 'Windows':
            _stdin.close()
        try:
            p.terminate()
            self.logger('terminating proxy: %s' % p, loglevel=log.loglevel_DEBUG)
        except OSError, e:
            if e.errno == 3:
                # No such process
                pass
        except WindowsError:
            pass

    def process_proxy_options(self):
        """\
        Override this method to incorporate elements from C{proxy_options} 
        into actual proxy subprocess execution.

        This method (if overridden) should (by design) never fail nor raise an exception.
        Make sure to catch all possible errors appropriately.

        If you want to log ignored proxy_options then

            1. remove processed proxy_options from self.proxy_options
            2. once you have finished processing the proxy_options call
            the parent class method X2GoProxyBASE.process_proxy_options()

        """
        # do the logging of remaining options
        if self.proxy_options:
            self.logger('ignoring non-processed proxy options: %s' % self.proxy_options, loglevel=log.loglevel_INFO)

    def _update_local_proxy_socket(self, port):
        pass

    def _generate_cmdline(self):
        return ''

    def start_proxy(self):
        """\
        Start the thread runner and wait for the proxy to come up.

        @return: a subprocess instance that knows about the externally started proxy command.
        @rtype: C{obj}

        """
        threading.Thread.start(self)

        # wait for proxy to get started
        _count = 0
        _maxwait = 40
        while self.proxy is None and _count < _maxwait:
            _count += 1
            self.logger('waiting for proxy to come up: 0.4s x %s' % _count, loglevel=log.loglevel_DEBUG)
            gevent.sleep(.4)

        if self.proxy:

            # also wait for fw_tunnel to become active
            _count = 0
            _maxwait = 40
            while (not self.fw_tunnel.is_active) and (not self.fw_tunnel.failed) and (_count < _maxwait):
                _count += 1
                self.logger('waiting for port fw tunnel to come up: 0.5s x %s' % _count, loglevel=log.loglevel_DEBUG)
                gevent.sleep(.5)

        return self.proxy, bool(self.proxy) and self.fw_tunnel.is_active

    def ok(self):
        """\
        Check if a proxy instance is up and running.

        @return: Proxy state, C{True} for proxy being up-and-running, C{False} otherwise
        @rtype C{bool}

        """
        return bool(self.proxy and self.proxy.poll() is None) and self.fw_tunnel.is_active



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
Print jobs can either be sent to any of the local print queues (CUPS, Win32API),
be opened in an external PDF viewer, be saved to a local folder or be handed 
over to a custom (print) command. This is defined by four print action classes
(L{X2GoPrintActionDIALOG}, L{X2GoPrintActionPDFVIEW}, L{X2GoPrintActionPDFSAVE}, L{X2GoPrintActionPRINT} and 
L{X2GoPrintActionPRINTCMD}).

"""
__NAME__ = 'x2goprintactions-pylib'

# modules
import os
import shutil
import copy
import time
import gevent

from defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS
if _X2GOCLIENT_OS in ("Windows"):
    import subprocess
    import win32api
    import win32print
else:
    import gevent_subprocess as subprocess
    import x2go_exceptions
    WindowsError = x2go_exceptions.WindowsError

# Python X2Go modules
import log
import defaults
# we hide the default values from epydoc (that's why we transform them to _UNDERSCORE variables)
import utils
import x2go_exceptions

_PRINT_ENV = os.environ.copy()


class X2GoPrintAction(object):

    __name__ = 'NAME'
    __description__ = 'DESCRIPTION'

    def __init__(self, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        This is a meta class and has no functionality as such. It is used as parent 
        class by »real« X2Go print actions.

        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoPrintAction} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        # these get set from within the X2GoPrintQueue class
        self.profile_name = 'UNKNOWN'
        self.session_name = 'UNKNOWN'

        self.client_instance = client_instance

    @property
    def name(self):
        """\
        Return the X2Go print action's name.

        """
        return self.__name__

    @property
    def description(self):
        """\
        Return the X2Go print action's description text.

        """
        return self.__description__

    def _do_print(self, pdf_file, job_title, spool_dir, ):
        """
        Perform the defined print action (doing nothing in L{X2GoPrintAction} parent class).

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param spool_dir: location of the X2Go client's spool directory
        @type spool_dir: C{str}

        """
        pass

    def do_print(self, pdf_file, job_title, spool_dir, ):
        """\
        Wrap around the actual print action (C{self._do_print}) with
        gevent.spawn().

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param spool_dir: location of the X2Go client's spool directory
        @type spool_dir: C{str}

        """
        pdf_file = os.path.normpath(pdf_file)
        spool_dir = os.path.normpath(spool_dir)

        self._do_print(pdf_file, job_title, spool_dir)

    def _humanreadable_filename(self, pdf_file, job_title, target_path):
        """\
        Extract a human readable filename for the X2Go print job file.

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param target_path: target path for human readable file
        @type target_path: C{str}
        @return: full readable file name path
        @rtype: C{str}

        """
        _hr_path = os.path.normpath(os.path.expanduser(os.path.join(os.path.normpath(target_path), '%s.pdf' % utils.slugify(job_title))))
        i = 0

        while os.path.exists(_hr_path):
            i += 1
            _hr_path = os.path.normpath(os.path.expanduser(os.path.join(os.path.normpath(target_path), '%s(%s).pdf' % (utils.slugify(job_title), i))))

        return _hr_path


class X2GoPrintActionPDFVIEW(X2GoPrintAction):
    """\
    Print action that views incoming print job in an external PDF viewer application.

    """
    __name__= 'PDFVIEW'
    __decription__= 'View as PDF document'

    pdfview_cmd = None

    def __init__(self, client_instance=None, pdfview_cmd=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param pdfview_cmd: command that starts the external PDF viewer application
        @type pdfview_cmd: C{str}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoPrintActionPDFVIEW} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        if pdfview_cmd is None:
            pdfview_cmd = defaults.DEFAULT_PDFVIEW_CMD
        self.pdfview_cmd = pdfview_cmd
        X2GoPrintAction.__init__(self, client_instance=client_instance, logger=logger, loglevel=loglevel)

    def _do_print(self, pdf_file, job_title, spool_dir, ):
        """\
        Open an incoming X2Go print job (PDF file) in an external PDF viewer application.

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param spool_dir: location of the X2Go client's spool directory
        @type spool_dir: C{str}

        @raise OSError: pass through all C{OSError}s except no. 2

        """
        pdf_file = os.path.normpath(pdf_file)
        spool_dir = os.path.normpath(spool_dir)

        if _X2GOCLIENT_OS == "Windows":
            self.logger('viewing incoming job in PDF viewer with Python\'s os.startfile( command): %s' % pdf_file, loglevel=log.loglevel_DEBUG)
            try:
                gevent.spawn(os.startfile, pdf_file)
            except WindowsError, win_err:
                if self.client_instance:
                    self.client_instance.HOOK_printaction_error(pdf_file,
                                                                profile_name=self.profile_name,
                                                                session_name=self.session_name,
                                                                err_msg=str(win_err)
                                                               )
                else:
                    self.logger('Encountered WindowsError: %s' % str(win_err), loglevel=log.loglevel_ERROR)
            time.sleep(20)
        else:
            _hr_filename = self._humanreadable_filename(pdf_file, job_title, spool_dir, )
            shutil.copy2(pdf_file, _hr_filename)
            cmd_line = [ self.pdfview_cmd, _hr_filename, ]
            self.logger('viewing incoming PDF with command: %s' % ' '.join(cmd_line), loglevel=log.loglevel_DEBUG)
            try:
                subprocess.Popen(cmd_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=_PRINT_ENV)
            except OSError, e:
                if e.errno == 2:
                    cmd_line = [ defaults.DEFAULT_PDFVIEW_CMD, _hr_filename ]
                    subprocess.Popen(cmd_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=_PRINT_ENV)
                else:
                    raise(e)
            self.logger('waiting 20s longer before deleting the PDF file ,,%s\'\'' % _hr_filename, loglevel=log.loglevel_DEBUG)
            time.sleep(20)
            os.remove(_hr_filename)


class X2GoPrintActionPDFSAVE(X2GoPrintAction):
    """\
    Print action that saves incoming print jobs to a local folder.

    """
    __name__ = 'PDFSAVE'
    __decription__= 'Save as PDF'

    save_to_folder = None

    def __init__(self, client_instance=None, save_to_folder=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param save_to_folder: saving location for incoming print jobs (PDF files)
        @type save_to_folder: C{str}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoPrintActionPDFSAVE} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        if save_to_folder is None:
            save_to_folder = defaults.DEFAULT_PDFSAVE_LOCATION
        if not utils.is_abs_path(save_to_folder):
            if not save_to_folder.startswith('~'):
                save_to_folder = os.path.normpath('~/%s' % save_to_folder)
            save_to_folder = os.path.expanduser(save_to_folder)
        self.save_to_folder = save_to_folder

        X2GoPrintAction.__init__(self, client_instance=client_instance, logger=None, loglevel=loglevel)

        self.logger('Save location for incoming PDFs is: %s' % self.save_to_folder, loglevel=log.loglevel_DEBUG)
        if not os.path.exists(self.save_to_folder):
            os.makedirs(self.save_to_folder, mode=0755)

    def _do_print(self, pdf_file, job_title, spool_dir):
        """\
        Save an incoming X2Go print job (PDF file) to a local folder.

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param spool_dir: location of the X2Go client's spool directory
        @type spool_dir: C{str}

        """
        pdf_file = os.path.normpath(pdf_file)
        spool_dir = os.path.normpath(spool_dir)

        dest_file = self._humanreadable_filename(pdf_file, job_title, target_path=self.save_to_folder)
        shutil.copy2(pdf_file, dest_file)


class X2GoPrintActionPRINT(X2GoPrintAction):
    """\
    Print action that actually prints an incoming print job file.

    """
    __name__ = 'PRINT'
    __decription__= 'UNIX/Win32GDI printing'

    def __init__(self, client_instance=None, printer=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param printer: name of the preferred printer, if C{None} the system's/user's default printer will be used
        @type printer: C{str}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoPrintActionPRINT} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        self.printer = printer
        X2GoPrintAction.__init__(self, client_instance=client_instance, logger=logger, loglevel=loglevel)

    def _do_print(self, pdf_file, job_title, spool_dir, ):
        """\
        Actually really print an incoming X2Go print job (PDF file) to a local printer device.

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param spool_dir: location of the X2Go client's spool directory
        @type spool_dir: C{str}

        """
        pdf_file = os.path.normpath(pdf_file)
        spool_dir = os.path.normpath(spool_dir)

        _hr_filename = self._humanreadable_filename(pdf_file, job_title, spool_dir)
        if _X2GOCLIENT_OS == 'Windows':
            _default_printer = win32print.GetDefaultPrinter()
            if self.printer:
                _printer = self.printer
                win32print.SetDefaultPrinter(_printer)
            else:
                _printer = _default_printer
            self.logger('printing incoming PDF file %s' % pdf_file, loglevel=log.loglevel_NOTICE)
            self.logger('printer name is ,,%s\'\'' % _printer, loglevel=log.loglevel_DEBUG)
            try:
                _stdin = file('nul', 'r')
                _shell = True
                if self.client_instance:
                    _gsprint_bin = self.client_instance.client_printing.get_value('print', 'gsprint')
                    self.logger('Using gsprint.exe path from printing config file: %s' % _gsprint_bin, loglevel=log.loglevel_DEBUG)
                else:
                    _program_files = os.environ['ProgramFiles']
                    _gsprint_bin = os.path.normpath(os.path.join(_program_files, 'ghostgum', 'gsview', 'gsprint.exe',))
                    self.logger('Using hard-coded gsprint.exe path: %s' % _gsprint_bin, loglevel=log.loglevel_DEBUG)
                self.logger('Trying Ghostgum tool ,,gsprint.exe'' for printing first (full path: %s)' % _gsprint_bin, loglevel=log.loglevel_DEBUG)
                subprocess.Popen([_gsprint_bin, pdf_file, ],
                                  stdin=_stdin,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  shell=_shell,
                                 )
                # give gsprint.exe a little time to find our printer
                time.sleep(10)

            except:
                self.logger('Falling back to win32api printing...', loglevel=log.loglevel_DEBUG)
                try:
                    win32api.ShellExecute (
                          0,
                          "print",
                          pdf_file,
                          None,
                          ".",
                          0
                    )
                    # give the win32api some time to find our printer...
                    time.sleep(10)
                except win32api.error, e:
                    if self.client_instance:
                        self.client_instance.HOOK_printaction_error(filename=_hr_filename, printer=_printer, err_msg=e.message, profile_name=self.profile_name, session_name=self.session_name)
                    else:
                        self.logger('Encountered win32api.error: %s' % str(e), loglevel=log.loglevel_ERROR)

            if self.printer:
                win32print.SetDefaultPrinter(_default_printer)
            time.sleep(60)

        else:
            _hr_filename = self._humanreadable_filename(pdf_file, job_title, spool_dir)
            self.logger('printing incoming PDF file %s' % _hr_filename, loglevel=log.loglevel_NOTICE)
            if self.printer:
                self.logger('printer name is %s' % self.printer, loglevel=log.loglevel_DEBUG)
            else:
                self.logger('using default CUPS printer', loglevel=log.loglevel_DEBUG)
            shutil.copy2(pdf_file, _hr_filename)
            if self.printer is None:
                cmd_line = [ 'lpr',
                             '-h',
                             '-r',
                             '-J%s' % job_title, 
                             '%s' % _hr_filename,
                           ]
            else:
                cmd_line = [ 'lpr',
                             '-h',
                             '-r',
                             '-P%s' % self.printer,
                             '-J%s' % job_title, 
                             '%s' % _hr_filename,
                           ]
            self.logger('executing local print command: %s' % " ".join(cmd_line), loglevel=log.loglevel_DEBUG)
            subprocess.Popen(cmd_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=_PRINT_ENV)

            # this is nasty!!!!
            self.logger('waiting 20s longer before deleting the PDF file ,,%s\'\'' % _hr_filename, loglevel=log.loglevel_DEBUG)
            time.sleep(20)
            try: os.remove(_hr_filename)
            except OSError: pass


class X2GoPrintActionPRINTCMD(X2GoPrintAction):
    """\
    Print action that calls an external command for further processing of incoming print jobs.

    The print job's PDF filename will be prepended as last argument to the print command
    used in L{X2GoPrintActionPRINTCMD} instances.

    """
    __name__      = 'PRINTCMD'
    __decription__= 'Print via a command (like LPR)'

    def __init__(self, client_instance=None, print_cmd=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param print_cmd: external command to be called on incoming print jobs
        @type print_cmd: C{str}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoPrintActionPRINTCMD} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        if print_cmd is None:
            print_cmd = defaults.DEFAULT_PRINTCMD_CMD
        self.print_cmd = print_cmd
        X2GoPrintAction.__init__(self, client_instance=client_instance, logger=logger, loglevel=loglevel)

    def _do_print(self, pdf_file, job_title, spool_dir):
        """\
        Execute an external command that has been defined on construction 
        of this L{X2GoPrintActionPRINTCMD} instance.

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param spool_dir: location of the X2Go client's spool directory
        @type spool_dir: C{str}

        """
        pdf_file = os.path.normpath(pdf_file)
        spool_dir = os.path.normpath(spool_dir)

        _hr_filename = self._humanreadable_filename(pdf_file, job_title, spool_dir)
        shutil.copy2(pdf_file, _hr_filename)
        self.logger('executing external command ,,%s\'\' on PDF file %s' % (self.print_cmd, _hr_filename), loglevel=log.loglevel_NOTICE)
        cmd_line = self.print_cmd.split()
        cmd_line.append(_hr_filename)
        self.logger('executing external command: %s' % " ".join(cmd_line), loglevel=log.loglevel_DEBUG)
        subprocess.Popen(cmd_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=_PRINT_ENV)

        # this is nasty!!!!
        self.logger('waiting 20s longer before deleting the PDF file ,,%s\'\'' % _hr_filename, loglevel=log.loglevel_DEBUG)
        time.sleep(20)
        try: os.remove(_hr_filename)
        except OSError: pass


class X2GoPrintActionDIALOG(X2GoPrintAction):
    """\
    Print action that mediates opening a print dialog window. This class is rather empty,
    the actual print dialog box must be implemented in our GUI application (with the application's
    L{X2GoClient} instance.

    """
    __name__      = 'DIALOG'
    __decription__= 'Open a print dialog box'

    def __init__(self, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: an L{X2GoClient} instance, within your customized L{X2GoClient} make sure 
            you have a C{HOOK_open_print_dialog(filename=<str>)} method defined that will actually
            open the print dialog.
        @type client_instance: C{obj}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoPrintActionDIALOG} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        @raise X2GoPrintActionException: if the client_instance has not been passed to the DIALOG print action

        """
        if client_instance is None:
            raise x2go_exceptions.X2GoPrintActionException('the DIALOG print action needs to know the X2GoClient instance (client=<instance>)')
        X2GoPrintAction.__init__(self, client_instance=client_instance, logger=logger, loglevel=loglevel)

    def _do_print(self, pdf_file, job_title, spool_dir):
        """\
        Execute an external command that has been defined on construction 
        of this L{X2GoPrintActionPRINTCMD} instance.

        @param pdf_file: PDF file name as placed in to the X2Go spool directory
        @type pdf_file: C{str}
        @param job_title: human readable print job title
        @type job_title: C{str}
        @param spool_dir: location of the X2Go client's spool directory
        @type spool_dir: C{str}

        """
        pdf_file = os.path.normpath(pdf_file)
        spool_dir = os.path.normpath(spool_dir)

        self.logger('Session %s (%s) is calling X2GoClient class hook method <client_instance>.HOOK_open_print_dialog' % (self.session_name, self.profile_name), loglevel=log.loglevel_NOTICE)
        _new_print_action = self.client_instance.HOOK_open_print_dialog(profile_name=self.profile_name, session_name=self.session_name)
        if _new_print_action and type(_new_print_action) != type(self):
            _new_print_action._do_print(pdf_file, job_title, spool_dir)




# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
L{X2GoSessionProfiles} class - managing x2goclient session profiles.

L{X2GoSessionProfiles} is a public API class. Use this class in your Python X2Go based 
applications.

"""
__NAME__ = 'x2gosessionprofiles-pylib'

# Python X2Go modules
from x2go.defaults import X2GO_SESSIONPROFILES_CONFIGFILES
from x2go.defaults import X2GO_SESSIONPROFILE_DEFAULTS
import x2go.inifiles as inifiles
import x2go.log as log

from x2go.x2go_exceptions import X2GoNotImplementedYetException

class X2GoSessionProfilesGCONF(inifiles.X2GoIniFile):

    defaultSessionProfile = X2GO_SESSIONPROFILE_DEFAULTS
    _non_profile_sections = ('embedded')

    def __init__(self, session_profile_defaults=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Retrieve X2Go session profiles from GConf daemon.

        @param session_profile_defaults: a default session profile
        @type session_profile_defaults: C{dict}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoSessionProfilesGCONF} constructor
        @type logger: L{X2GoLogger} instance
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        raise X2GoNotImplementedYetException('GCONF backend support is not implemented yet')



# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2Go reverse SSH/Paramiko tunneling provides X2Go sound, X2Go printing and
X2Go sshfs for folder sharing and mounting remote devices in X2Go terminal
server sessions.

"""
__NAME__ = 'x2gorevtunnel-pylib'

# modules
import copy
import threading
import gevent
import paramiko

# gevent/greenlet
from gevent import select, socket, Timeout

# Python X2Go modules
import log


def x2go_transport_tcp_handler(chan, (origin_addr, origin_port), (server_addr, server_port)):
    """\
    An X2Go customized TCP handler for the Paramiko/SSH C{Transport()} class.

    Incoming channels will be put into Paramiko's default accept queue. This corresponds to 
    the default behaviour of Paramiko's C{Transport} class.

    However, additionally this handler function checks the server port of the incoming channel
    and detects if there are Paramiko/SSH reverse forwarding tunnels waiting for the incoming 
    channels. The Paramiko/SSH reverse forwarding tunnels are initiated by an L{X2GoSession} instance
    (currently supported: reverse tunneling auf audio data, reverse tunneling of SSH requests).

    If the server port of an incoming Paramiko/SSH channel matches the configured port of an L{X2GoRevFwTunnel} 
    instance, this instance gets notified of the incoming channel and a new L{X2GoRevFwChannelThread} is 
    started. This L{X2GoRevFwChannelThread} then takes care of the new channel's incoming data stream.

    """
    transport = chan.get_transport()
    transport._queue_incoming_channel(chan)
    rev_tuns = transport.reverse_tunnels

    for session_name in rev_tuns.keys():

        if int(server_port) in [ int(tunnel[0]) for tunnel in rev_tuns[session_name].values() ]:

            if rev_tuns[session_name]['snd'] is not None and int(server_port) == int(rev_tuns[session_name]['snd'][0]):
                rev_tuns[session_name]['snd'][1].notify()

            elif rev_tuns[session_name]['sshfs'] is not None and int(server_port) == int(rev_tuns[session_name]['sshfs'][0]):
                rev_tuns[session_name]['sshfs'][1].notify()


class X2GoRevFwTunnel(threading.Thread):
    """\
    L{X2GoRevFwTunnel} class objects are used to reversely tunnel 
    X2Go audio, X2Go printing and X2Go folder sharing / device mounting
    through Paramiko/SSH.

    """
    def __init__(self, server_port, remote_host, remote_port, ssh_transport, session_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Setup a reverse tunnel through Paramiko/SSH.

        After the reverse tunnel has been setup up with L{X2GoRevFwTunnel.start()} it waits
        for notification from L{X2GoRevFwTunnel.notify()} to accept incoming channels. This 
        notification (L{X2GoRevFwTunnel.notify()} gets called from within the transport's 
        TCP handler function L{x2go_transport_tcp_handler} of the L{X2GoSession} instance.

        @param server_port: the TCP/IP port on the X2Go server (starting point of the tunnel), 
            normally some number above 30000
        @type server_port: int
        @param remote_host: the target address for reversely tunneled traffic. With X2Go this should 
            always be set to the localhost (IPv4) address.
        @type remote_host: str
        @param remote_port: the TCP/IP port on the X2Go client (end point of the tunnel),
            normally an application's standard port (22 for SSH, 4713 for pulse audio, etc.)
        @type remote_port: int
        @param ssh_transport: the L{X2GoSession}'s Paramiko/SSH transport instance
        @type ssh_transport: C{paramiko.Transport} instance
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoRevFwTunnel} constructor
        @type logger: L{X2GoLogger} instance
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: int

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.server_port = server_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.ssh_transport = ssh_transport
        self.session_instance = session_instance

        self.open_channels = {}
        self.incoming_channel = threading.Condition()

        threading.Thread.__init__(self)
        self.daemon = True
        self._accept_channels = True

    def __del__(self):
        """\
        Class destructor.

        """
        self.stop_thread()
        self.cancel_port_forward('', self.server_port)

    def cancel_port_forward(self, address, port):
        """\
        Cancel a port forwarding request. This cancellation request is sent to the server and
        on the server the port forwarding should be unregistered.

        @param address: remote server address
        @type address: C{str}
        @param port: remote port
        @type port: C{int}

        """
        timeout = Timeout(10)
        timeout.start()
        try:
            self.ssh_transport.global_request('cancel-tcpip-forward', (address, port), wait=True)
        except:
            pass
        finally:
            timeout.cancel()

    def pause(self):
        """\
        Prevent acceptance of new incoming connections through the Paramiko/SSH
        reverse forwarding tunnel. Also, any active connection on this L{X2GoRevFwTunnel}
        instance will be closed immediately, if this method is called.

        """
        if self._accept_channels == True:
            self.cancel_port_forward('', self.server_port)
            self._accept_channels = False
            self.logger('paused thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

    def resume(self):
        """\
        Resume operation of the Paramiko/SSH reverse forwarding tunnel
        and continue accepting new incoming connections.

        """
        if self._accept_channels == False:
            self._accept_channels = True
            self._requested_port = self.ssh_transport.request_port_forward('', self.server_port, handler=x2go_transport_tcp_handler)
            self.logger('resumed thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)

    def notify(self):
        """\
        Notify an L{X2GoRevFwTunnel} instance of an incoming Paramiko/SSH channel.

        If an incoming reverse tunnel channel appropriate for this instance has
        been detected, this method gets called from the L{X2GoSession}'s transport
        TCP handler.

        The sent notification will trigger a C{thread.Condition()} waiting for notification
        in L{X2GoRevFwTunnel.run()}.

        """
        self.incoming_channel.acquire()
        self.logger('notifying thread of incoming channel: %s' % repr(self), loglevel=log.loglevel_DEBUG)
        self.incoming_channel.notify()
        self.incoming_channel.release()

    def stop_thread(self):
        """\
        Stops this L{X2GoRevFwTunnel} thread completely.

        """
        self.pause()
        self._keepalive = False
        self.logger('stopping thread: %s' % repr(self), loglevel=log.loglevel_DEBUG)
        self.notify()

    def _request_port_forwarding(self):
        try:
            self._requested_port = self.ssh_transport.request_port_forward('', self.server_port, handler=x2go_transport_tcp_handler)
        except paramiko.SSHException:
            # if port forward request fails, we try to tell the server to cancel all foregoing port forward requests on 
            # self.server_port
            self.cancel_port_forward('', self.server_port)
            gevent.sleep(1)
            try:
                self._requested_port = self.ssh_transport.request_port_forward('', self.server_port, handler=x2go_transport_tcp_handler)
            except paramiko.SSHException, e:
                if self.session_instance:
                    self.session_instance.HOOK_rforward_request_denied(server_port=self.server_port)
                else:
                    self.logger('Encountered SSHException: %s (for reverse TCP port forward with local destination port %s' % (str(e), self.server_port), loglevel=log.loglevel_WARN)

    def run(self):
        """\
        This method gets run once an L{X2GoRevFwTunnel} has been started with its
        L{start()} method. Use L{X2GoRevFwTunnel}.stop_thread() to stop the
        reverse forwarding tunnel again. You can also temporarily lock the tunnel
        down with L{X2GoRevFwTunnel.pause()} and L{X2GoRevFwTunnel.resume()}).

        L{X2GoRevFwTunnel.run()} waits for notifications of an appropriate incoming
        Paramiko/SSH channel (issued by L{X2GoRevFwTunnel.notify()}). Appropriate in
        this context means, that its start point on the X2Go server matches the class's
        property C{server_port}.

        Once a new incoming channel gets announced by the L{notify()} method, a new 
        L{X2GoRevFwChannelThread} instance will be initialized. As a data stream handler,
        the function L{x2go_rev_forward_channel_handler()} will be used.

        The channel will last till the connection gets dropped on the X2Go server side or 
        until the tunnel gets paused by an L{X2GoRevFwTunnel.pause()} call or stopped via the
        L{X2GoRevFwTunnel.stop_thread()} method.

        """
        self._request_port_forwarding()
        self._keepalive = True
        while self._keepalive:

            self.incoming_channel.acquire()

            self.logger('waiting for incoming data channel on X2Go server port: [127.0.0.1]:%s' % self.server_port, loglevel=log.loglevel_DEBUG)
            self.incoming_channel.wait()

            if self._keepalive:
                self.logger('detected incoming data channel on X2Go server port: [127.0.0.1]:%s' % self.server_port, loglevel=log.loglevel_DEBUG)
                _chan = self.ssh_transport.accept()
                self.logger('data channel %s for server port [127.0.0.1]:%s is up' % (_chan, self.server_port), loglevel=log.loglevel_DEBUG)
            else:
                self.logger('closing down rev forwarding tunnel on remote end [127.0.0.1]:%s' % self.server_port, loglevel=log.loglevel_DEBUG)

            self.incoming_channel.release()
            if self._accept_channels and self._keepalive:
                _new_chan_thread = X2GoRevFwChannelThread(_chan, (self.remote_host, self.remote_port), 
                                                          target=x2go_rev_forward_channel_handler, 
                                                          kwargs={ 
                                                            'chan': _chan,
                                                            'addr': self.remote_host,
                                                            'port': self.remote_port,
                                                            'parent_thread': self, 
                                                            'logger': self.logger, 
                                                          }
                                                  )
                _new_chan_thread.start()
                self.open_channels['[%s]:%s' % _chan.origin_addr] = _new_chan_thread


def x2go_rev_forward_channel_handler(chan=None, addr='', port=0, parent_thread=None, logger=None, ):
    """\
    Handle the data stream of a requested channel that got set up by a L{X2GoRevFwTunnel} (Paramiko/SSH 
    reverse forwarding tunnel).

    The channel (and the corresponding connections) close either ...

        - ... if the connecting application closes the connection and thus, drops 
        the channel, or
        - ... if the L{X2GoRevFwTunnel} parent thread gets paused. The call
        of L{X2GoRevFwTunnel.pause()} on the instance can be used to shut down all incoming 
        tunneled SSH connections associated to this L{X2GoRevFwTunnel} instance
        from within a Python X2Go application.

    @param chan: channel
    @type chan: C{class}
    @param addr: bind address
    @type addr: C{str}
    @param port: bind port
    @type port: C{int}
    @param parent_thread: the calling L{X2GoRevFwTunnel} instance
    @type parent_thread: L{X2GoRevFwTunnel} instance
    @param logger: you can pass an L{X2GoLogger} object to the
        L{X2GoRevFwTunnel} constructor
    @type logger: L{X2GoLogger} instance

    """
    fw_socket = socket.socket()
    fw_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    if logger is None:
        def _dummy_logger(msg, l):
            pass
        logger = _dummy_logger

    try:
        fw_socket.connect((addr, port))
    except Exception, e:
        logger('Reverse forwarding request to %s:%d failed: %r' % (addr, port, e), loglevel=log.loglevel_INFO)
        return

    logger('Connected! Reverse tunnel open %r -> %r -> %r' % (chan.origin_addr,
                                                              chan.getpeername(), (addr, port)), 
                                                              loglevel=log.loglevel_INFO)
    while parent_thread._accept_channels:
        r, w, x = select.select([fw_socket, chan], [], [])
        if fw_socket in r:
            data = fw_socket.recv(1024)
            if len(data) == 0:
                break
            chan.send(data)
        if chan in r:
            data = chan.recv(1024)
            if len(data) == 0:
                break
            fw_socket.send(data)

    chan.close()
    fw_socket.close()
    logger('Reverse tunnel %s closed from %r' % (chan, chan.origin_addr,), loglevel=log.loglevel_INFO)


class X2GoRevFwChannelThread(threading.Thread):
    """\
    Starts a thread for each incoming Paramiko/SSH data channel trough the reverse
    forwarding tunnel.

    """
    def __init__(self, channel, remote=None, **kwargs):
        """\
        Initializes a reverse forwarding channel thread.

        @param channel: incoming Paramiko/SSH channel from the L{X2GoSession}'s transport
            accept queue
        @type channel: class
        @param remote: tuple (addr, port) that specifies the data endpoint of the channel
        @type remote: C{tuple(str, int)}

        """
        self.channel = channel
        if remote is not None:
            self.remote_host = remote[0]
            self.remote_port = remote[1]
        threading.Thread.__init__(self, **kwargs)
        self.daemon = True



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoTerminalSession class - core functions for handling your individual X2Go sessions.

This backend handles X2Go server implementations that respond with session infos 
via server-side STDOUT and use NX3 as graphical proxy.

"""
__NAME__ = 'x2goterminalsession-pylib'

# modules
import os
import types
import gevent
import cStringIO
import copy
import shutil
import threading

# Python X2Go modules
import x2go.rforward as rforward
import x2go.sftpserver as sftpserver
import x2go.printqueue as printqueue
import x2go.mimebox as mimebox
import x2go.log as log
import x2go.defaults as defaults
import x2go.utils as utils
import x2go.x2go_exceptions as x2go_exceptions

# we hide the default values from epydoc (that's why we transform them to _UNDERSCORE variables)
from x2go.defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS
from x2go.defaults import LOCAL_HOME as _LOCAL_HOME
from x2go.defaults import CURRENT_LOCAL_USER as _CURRENT_LOCAL_USER
from x2go.defaults import X2GO_CLIENT_ROOTDIR as _X2GO_CLIENT_ROOTDIR
from x2go.defaults import X2GO_SESSIONS_ROOTDIR as _X2GO_SESSIONS_ROOTDIR
from x2go.defaults import X2GO_GENERIC_APPLICATIONS as _X2GO_GENERIC_APPLICATIONS
from x2go.defaults import X2GO_DESKTOPSESSIONS as _X2GO_DESKTOPSESSIONS

from x2go.backends.info import X2GoServerSessionInfo as _X2GoServerSessionInfo
from x2go.backends.info import X2GoServerSessionList as _X2GoServerSessionList
from x2go.backends.proxy import X2GoProxy as _X2GoProxy
from x2go.backends.printing import X2GoClientPrinting as _X2GoClientPrinting

_local_color_depth = utils.local_color_depth()

def _rewrite_cmd(cmd, params=None):
    """\
    Mechansim that rewrites X2Go server commands into something that gets understood by
    the server-side script C{x2goruncommand}.

    @param cmd: the current command for execution (as found in the session profile parameter C{cmd})
    @type cmd: C{str}
    @param params: an session paramter object
    @type params: L{X2GoSessionParams}

    @return: the rewritten command for server-side execution
    @rtype: C{str}

    """
    # start with an empty string
    cmd = cmd or ''

    # find window manager commands
    if cmd in _X2GO_DESKTOPSESSIONS.keys():
        cmd = _X2GO_DESKTOPSESSIONS[cmd]

    if (cmd == 'RDP') and (type(params) == X2GoSessionParams):
        _depth = params.depth
        if int(_depth) == 17:
            _depth = 16
        if params.geometry == 'fullscreen':
            cmd = 'rdesktop -f -N %s %s -a %s' % (params.rdp_options, params.rdp_server, _depth)
        else:
            cmd = 'rdesktop -g %s -N %s %s -a %s' % (params.geometry, params.rdp_options, params.rdp_server, _depth)

    # place quot marks around cmd if not empty string
    if cmd:
        cmd = '"%s"' % cmd

    if ((type(params) == X2GoSessionParams) and params.published_applications and cmd == ''):
        cmd = 'PUBLISHED'

    return cmd


def _rewrite_blanks(cmd):
    """\
    In command strings X2Go server scripts expect blanks being rewritten to ,,X2GO_SPACE_CHAR''.

    @param cmd: command that has to be rewritten for passing to the server
    @type cmd: C{str}

    @return: the command with blanks rewritten to ,,X2GO_SPACE_CHAR''
    @rtype: C{str}

    """
    # X2Go run command replace X2GO_SPACE_CHAR string with blanks
    if cmd:
        cmd = cmd.replace(" ", "X2GO_SPACE_CHAR")
    return cmd


class X2GoSessionParams(object):
    """\
    The L{X2GoSessionParams} class is used to store all parameters that
    C{X2GoTerminalSession} backend objects are constructed with.

    """
    def rewrite_session_type(self):
        """\
        Rewrite the X2Go session type, so that the X2Go server
        can understand it (C{desktop} -> C{D}, etc.).

        Also if the object's C{command} property is a known window 
        manager, the session type will be set to 'D' 
        (i.e. desktop).

        @return: 'D' if session should probably a desktop session,
            'R' for rootless sessions, 'P' for sessions providing published
            applications, and 'S' for desktop sharing sessions
        @rtype: C{str}

        """
        cmd = self.cmd
        published = self.published_applications

        if published and self.cmd in ('', 'PUBLISHED'):
            self.session_type = 'P'
            self.cmd = 'PUBLISHED'
        else:
            if cmd == 'RDP' or cmd.startswith('rdesktop') or cmd.startswith('xfreedrp'):
                if self.geometry == 'fullscreen': self.session_type = 'D'
                else: self.session_type = 'R'
            elif cmd == 'XDMCP':
                self.session_type = 'D'
            elif cmd in _X2GO_DESKTOPSESSIONS.keys():
                self.session_type = 'D'
            elif os.path.basename(cmd) in _X2GO_DESKTOPSESSIONS.values():
                self.session_type = 'D'

        if self.session_type in ("D", "desktop"):
            self.session_type = 'D'
        elif self.session_type in ("S", "shared", "shadow"):
            self.session_type = 'S'
        elif self.session_type in ("R", "rootless", "application"):
            self.session_type = 'R'
        elif self.session_type in ("P", "published", "published_applications"):
            self.session_type = 'P'

        return self.session_type

    def update(self, **properties_to_be_updated):
        """\
        Update all properties in the object L{X2GoSessionParams} object from
        the passed on dictionary.

        @param properties_to_be_updated: a dictionary with L{X2GoSessionParams}
            property names as keys und their values to be update in 
            L{X2GoSessionParams} object.
        @type properties_to_be_updated: C{dict}

        """
        for key in properties_to_be_updated.keys():
            setattr(self, key, properties_to_be_updated[key] or '')
        self.rewrite_session_type()


class X2GoTerminalSessionSTDOUT(object):
    """\
    Class for managing X2Go terminal sessions on a remote X2Go server via Paramiko/SSH.

    With the L{X2GoTerminalSessionSTDOUT} class you can start new X2Go sessions, resume suspended 
    sessions or suspend resp. terminate currently running sessions on a 
    connected X2Go server.

    An L{X2GoTerminalSessionSTDOUT} object uses two main data structure classes:

        - L{X2GoSessionParams}: stores all parameters that have been passed to the 
        constructor method.

        - C{X2GoServerSessionInfo*} backend class: when starting or resuming a session, an object of this class 
        will be used to store all information retrieved from the X2Go server.

    The terminal session instance works closely together (i.e. depends on) a connected control
    session instance (e.g. L{X2GoControlSessionSTDOUT}). You never should use either of them as a standalone
    instance. Both, control session and terminal session(s) get managed/controlled via L{X2GoSession} instances.

    """
    def __init__(self, control_session, session_info=None,
                 geometry="800x600", depth=_local_color_depth, link="adsl", pack="16m-jpeg-9", dpi='',
                 cache_type="unix-kde",
                 kbtype='null/null', kblayout='null', kbvariant='null',
                 session_type="application", snd_system='pulse', snd_port=4713, cmd=None,
                 published_applications=False,
                 set_session_title=False, session_title="", applications=[],
                 rdp_server=None, rdp_options=None,
                 xdmcp_server=None,
                 convert_encoding=False, server_encoding='UTF-8', client_encoding='UTF-8',
                 rootdir=None,
                 profile_name='UNKNOWN', profile_id=utils._genSessionProfileId(),
                 print_action=None, print_action_args={},
                 info_backend=_X2GoServerSessionInfo,
                 list_backend=_X2GoServerSessionList,
                 proxy_backend=_X2GoProxy, proxy_options={},
                 printing_backend=_X2GoClientPrinting,
                 client_rootdir=os.path.join(_LOCAL_HOME, _X2GO_CLIENT_ROOTDIR),
                 sessions_rootdir=os.path.join(_LOCAL_HOME, _X2GO_SESSIONS_ROOTDIR),
                 session_instance=None,
                 logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Initialize an X2Go session. With the L{X2GoTerminalSessionSTDOUT} class you can start
        new X2Go sessions, resume suspended sessions or suspend resp. terminate
        currently running sessions on a connected X2Go server.

        @param geometry: screen geometry of the X2Go session. Can be either C{<width>x<height>},
            C{maximize} or C{fullscreen}
        @type geometry: C{str}
        @param depth: color depth in bits (common values: C{16}, C{24})
        @type depth: C{int}
        @param link: network link quality (either one of C{modem}, C{isdn}, C{adsl}, C{wan} or C{lan})
        @type link: C{str}
        @param pack: compression method for NX based session proxying
        @type pack: C{str}
        @param dpi: dots-per-inch value for the session screen (has an impact on the font size on screen)
        @type dpi: C{str}
        @param cache_type: a dummy parameter that is passed to the L{X2GoProxyBASE}. In NX Proxy 
            (class C{X2GoProxyNX3}) this originally is the session name. With X2Go it 
            defines the name of the NX cache directory. Best is to leave it untouched.
        @type cache_type: C{str}
        @param kbtype: keyboard type, e.g. C{pc105/us} (default), C{pc105/de}, ...
        @type kbtype: C{str}
        @param kblayout: keyboard layout, e.g. C{us} (default), C{de}, C{fr}, ...
        @type kblayout: C{str}
        @param kbvariant: keyboard variant, e.g. C{nodeadkeys} (for C{de} layout), C{intl} (for C{us} layout), etc.
        @type kbvariant: C{str}
        @param session_type: either C{desktop}, C{application} (rootless session) or C{shared}
        @type session_type: C{str}
        @param snd_system: sound system to be used on server (C{none}, C{pulse} (default), 
            C{arts} (obsolete) or C{esd})
        @type snd_system: C{str}
        @param snd_port: local sound port for network capable audio system
        @type snd_port: C{int}
        @param cmd: command to be run on X2Go server after session start (only used
            when L{X2GoTerminalSessionSTDOUT.start()} is called, ignored on resume, suspend etc.
        @type cmd: C{str}
        @param published_applications: session is published applications provider
        @type published_applications: C{bool}
        @param set_session_title: modify the session title (i.e. the Window title) of desktop or shared desktop sessions
        @type set_session_title: C{bool}
        @param session_title: session title for this (desktop or shared desktop) session
        @type session_title: C{str}
        @param applications: applications available for rootless application execution
        @type applications: C{list}
        @param rdp_server: host name of server-side RDP server
        @type rdp_server: C{str}
        @param rdp_options: options for the C{rdesktop} command executed on the X2Go server (RDP proxy mode of X2Go)
        @type rdp_options: C{str}
        @param xdmcp_server: XDMCP server to connect to
        @type xdmcp_server: C{str}
        @param convert_encoding: convert file system encodings between server and client (for client-side shared folders)
        @type convert_encoding: C{bool}
        @param server_encoding: server-side file system / session encoding
        @type server_encoding: C{str}
        @param client_encoding: client-side file system encoding (if client-side is MS Windows, this parameter gets overwritten to WINDOWS-1252)
        @type client_encoding: C{str}
        @param rootdir: X2Go session directory, normally C{~/.x2go}
        @type rootdir: C{str}
        @param profile_name: the session profile name for this terminal session
        @type profile_name: C{str}
        @param profile_id: the session profile ID for this terminal session
        @type profile_id: C{str}
        @param print_action: either a print action short name (PDFVIEW, PDFSAVE, PRINT, PRINTCMD) or the
            resp. C{X2GoPrintActionXXX} class (where XXX equals one of the given short names)
        @type print_action: C{str} or C{class}
        @param print_action_args: optional arguments for a given print_action (for further info refer to
            L{X2GoPrintActionPDFVIEW}, L{X2GoPrintActionPDFSAVE}, L{X2GoPrintActionPRINT} and L{X2GoPrintActionPRINTCMD})
        @type print_action_args: dict
        @param info_backend: backend for handling storage of server session information
        @type info_backend: C{X2GoServerSessionInfo*} instance
        @param list_backend: backend for handling storage of session list information
        @type list_backend: C{X2GoServerSessionList*} instance
        @param proxy_backend: backend for handling the X-proxy connections
        @type proxy_backend: C{X2GoProxy*} instance
        @param proxy_options: a set of very C{X2GoProxy*} backend specific options; any option that is not known
            to the C{X2GoProxy*} backend will simply be ignored
        @type proxy_options: C{dict}
        @param client_rootdir: client base dir (default: ~/.x2goclient)
        @type client_rootdir: C{str}
        @param sessions_rootdir: sessions base dir (default: ~/.x2go)
        @type sessions_rootdir: C{str}
        @param session_instance: the L{X2GoSession} instance that is parent to this terminal session
        @type session_instance: C{obj}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoTerminalSessionSTDOUT} constructor
        @type logger: L{X2GoLogger} instance
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        self.proxy = None
        self.proxy_subprocess = None
        self.proxy_options = proxy_options

        self.active_threads = []
        self.reverse_tunnels = {}

        self.print_queue = None
        self.mimebox_queue = None

        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.control_session = control_session
        self.reverse_tunnels = self.control_session.get_transport().reverse_tunnels

        self.client_rootdir = client_rootdir
        self.sessions_rootdir = sessions_rootdir

        self.params = X2GoSessionParams()

        self.params.geometry = str(geometry)
        self.params.link = str(link)
        self.params.pack = str(pack)
        self.params.dpi = str(dpi)
        self.params.cache_type = str(cache_type)
        self.params.session_type = str(session_type)
        self.params.kbtype = str(kbtype)
        self.params.kblayout = str(kblayout)
        self.params.kbvariant = str(kbvariant)
        self.params.snd_system = str(snd_system)
        self.params.cmd = str(cmd)
        self.params.depth = str(depth)

        self.params.published_applications = published_applications
        self.published_applications = published_applications

        self.params.rdp_server = str(rdp_server)
        self.params.rdp_options = str(rdp_options)
        self.params.xdmcp_server = str(xdmcp_server)

        self.params.convert_encoding = convert_encoding
        self.params.client_encoding = str(client_encoding)
        self.params.server_encoding = str(server_encoding)

        self.params.rootdir = (type(rootdir) is types.StringType) and rootdir or self.sessions_rootdir
        self.params.update()

        self.profile_name = profile_name
        self.set_session_title = set_session_title
        self.session_title = session_title
        self.session_window = None
        self.proxy_backend = proxy_backend

        self.snd_port = snd_port
        self.print_action = print_action
        self.print_action_args = print_action_args
        self.printing_backend = printing_backend
        self.session_instance = session_instance
        if self.session_instance:
            self.client_instance = self.session_instance.client_instance
        else:
            self.client_instance = None

        self._share_local_folder_busy = False
        self._mk_sessions_rootdir(self.params.rootdir)

        self.session_info = session_info
        if self.session_info is not None:
            if self.session_info.name:
                self.session_info.local_container = os.path.join(self.params.rootdir, 'S-%s' % self.session_info.name)
            else:
                raise x2go_exceptions.X2GoTerminalSessionException('no valid session info availble')
        else:
            self.session_info = info_backend()

        self._share_local_folder_lock = threading.Lock()
        self._cleaned_up = False

    def __del__(self):
        """\
        Tidy up if terminal session gets destructed.

        """
        self._x2go_tidy_up()

    def _x2go_tidy_up(self):
        """\
        Tidy up this terminal session...
          - shutdown all forwarding and reverse forwarding tunnels
          - shutdown the print queue (if running)
          - shutdown the MIME box queue (if running)
          - clear the session info

        """
        self.release_proxy()

        try:

            if self.control_session.get_transport() is not None:
                try:
                    for _tunnel in [ _tun[1] for _tun in self.reverse_tunnels[self.session_info.name].values() ]:
                        if _tunnel is not None:
                            _tunnel.__del__()
                except KeyError:
                    pass

            if self.print_queue is not None:
                self.print_queue.__del__()

            if self.mimebox_queue is not None:
                self.mimebox_queue.__del__()

        except AttributeError:
            pass

        self.session_info.clear()

    def _mk_sessions_rootdir(self, rootdir):
        """\
        Create the server-side session root dir (normally ~/.x2go).

        @param rootdir: server-side session root directory
        @type rootdir: C{str}

        """
        try:
            os.makedirs(rootdir)
        except OSError, e:
            if e.errno == 17:
                # file exists
                pass
            else:
                raise OSError, e

    def _rm_session_dirtree(self):
        """\
        Purge client-side session dir (session cache directory).

        """
        if self.session_info.name:
            shutil.rmtree('%s/S-%s' % (self.params.rootdir, self.session_info), ignore_errors=True)

    def _rm_desktop_dirtree(self):
        """\
        Purge client-side session dir (C-<display> directory)

        """
        if self.session_info.display:
            shutil.rmtree('%s/S-%s' % (self.params.rootdir, self.session_info.display), ignore_errors=True)

    def get_session_name(self):
        """\
        Retrieve the X2Go session's name from the session info object.

        @return: the session name
        @rtype: C{str}

        """
        return self.session_info.name

    def get_session_info(self):
        """\
        Retrieve the X2Go session's session info object.

        @return: the session info object
        @rtype: C{X2GoServerSessionInfo*}

        """
        return self.session_info

    def get_session_cmd(self):
        """\
        Retrieve the X2Go session's command as stored in the session parameter object.

        @return: the session command
        @rtype: C{str}

        """
        return self.params.cmd

    def get_session_type(self):
        """\
        Retrieve the X2Go session's session type as stored in the session parameter object.

        @return: the session type
        @rtype: C{str}

        """
        return self.params.session_type

    def start_sound(self):
        """\
        Initialize Paramiko/SSH reverse forwarding tunnel for X2Go sound.

        Currently supported audio protocols:

            - PulseAudio
            - Esound (not tested very much)

        @raise X2GoControlSessionException: if the control session of this terminal session is not connected

        """
        _tunnel = None
        if self.reverse_tunnels[self.session_info.name]['snd'][1] is None:
            if self.params.snd_system == 'pulse':
                self.logger('initializing PulseAudio sound support in X2Go session', loglevel=log.loglevel_INFO)
                ###
                ### PULSEAUDIO
                ###
                if os.path.exists(os.path.normpath('%s/.pulse-cookie' % _LOCAL_HOME)):
                    # setup pulse client config file on X2Go server
                    cmd_line = "echo 'default-server=127.0.0.1:%s'>%s/.pulse-client.conf;" % (self.session_info.snd_port, self.session_info.remote_container) + \
                               "echo 'cookie-file=%s/.pulse-cookie'>>%s/.pulse-client.conf" % (self.session_info.remote_container, self.session_info.remote_container)
                    (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)

                    self.control_session._x2go_sftp_put(local_path='%s/.pulse-cookie' % _LOCAL_HOME, remote_path='%s/.pulse-cookie' % self.session_info.remote_container)

                    # start reverse SSH tunnel for pulse stream
                    _tunnel = rforward.X2GoRevFwTunnel(server_port=self.session_info.snd_port, 
                                                       remote_host='127.0.0.1', 
                                                       remote_port=self.snd_port, 
                                                       ssh_transport=self.control_session.get_transport(),
                                                       session_instance=self.session_instance,
                                                       logger=self.logger
                                                      )
                else:
                    if self.client_instance:
                        self.client_instance.HOOK_on_sound_tunnel_failed(profile_name=self.profile_name, session_name=self.session_info.name)
            elif self.params.snd_system == 'arts':
                ###
                ### ARTSD AUDIO
                ###
                self.logger('the ArtsD sound server (as in KDE3) is obsolete and will not be supported by Python X2Go...', loglevel=log.loglevel_WARNING)

            elif self.params.snd_system == 'esd':
                ###
                ### ESD AUDIO
                ###

                self.logger('initializing ESD sound support in X2Go session', loglevel=log.loglevel_INFO)
                self.control_session._x2go_sftp_put(local_path='%s/.esd_auth' % _LOCAL_HOME, remote_path='%s/.esd_auth' % self.control_session._x2go_remote_home)

                # start reverse SSH tunnel for pulse stream
                _tunnel = rforward.X2GoRevFwTunnel(server_port=self.session_info.snd_port, 
                                                   remote_host='127.0.0.1', 
                                                   remote_port=self.snd_port, 
                                                   ssh_transport=self.control_session.get_transport(),
                                                   session_instance=self.session_instance,
                                                   logger=self.logger
                                                  )


            if _tunnel is not None:
                self.reverse_tunnels[self.session_info.name]['snd'] = (self.session_info.snd_port, _tunnel)
                _tunnel.start()
                self.active_threads.append(_tunnel)

        else:
            # tunnel has already been started and might simply need a resume call
            self.reverse_tunnels[self.session_info.name]['snd'][1].resume()

    def start_sshfs(self):
        """\
        Initialize Paramiko/SSH reverse forwarding tunnel for X2Go folder sharing.

        """
        if not self.control_session.is_sshfs_available():
            raise x2go_exceptions.X2GoUserException('Remote user %s is not allowed to share SSHFS resources with the server.' % self.session_info.username)

        # start reverse SSH tunnel for sshfs (folder sharing, printing)
        ssh_transport = self.control_session.get_transport()
        if self.reverse_tunnels[self.session_info.name]['sshfs'][1] is None:

            _tunnel = sftpserver.X2GoRevFwTunnelToSFTP(server_port=self.session_info.sshfs_port,
                                                       ssh_transport=ssh_transport,
                                                       auth_key=self.control_session._x2go_session_auth_rsakey,
                                                       session_instance=self.session_instance,
                                                       logger=self.logger
                                                      )

            if _tunnel is not None:
                self.reverse_tunnels[self.session_info.name]['sshfs'] = (self.session_info.sshfs_port, _tunnel)
                _tunnel.start()
                self.active_threads.append(_tunnel)
                while not _tunnel.ready:
                    gevent.sleep(.1)

        else:
            # tunnel has already been started and might simply need a resume call
            self.reverse_tunnels[self.session_info.name]['sshfs'][1].resume()

    def _x2go_pause_rev_fw_tunnel(self, name):
        """\
        Pause reverse SSH tunnel of name <name>.

        @param name: tunnel name (either of C{sshfs}, C{snd})
        @type name: C{str}

        """
        _tunnel = self.reverse_tunnels[self.session_info.name][name][1]
        if _tunnel is not None:
            _tunnel.pause()

    def stop_sound(self):
        """\
        Shutdown (pause) Paramiko/SSH reverse forwarding tunnel for X2Go sound.

        """
        self._x2go_pause_rev_fw_tunnel('snd')

    def stop_sshfs(self):
        """\
        Shutdown (pause) Paramiko/SSH reverse forwarding tunnel for X2Go folder sharing.

        """
        self._x2go_pause_rev_fw_tunnel('sshfs')

    def start_printing(self):
        """\
        Initialize X2Go print spooling.

        @raise X2GoUserException: if the X2Go printing feature is not available to this user

        """
        if not self.control_session.is_sshfs_available():
            raise x2go_exceptions.X2GoUserException('Remote user %s is not allowed to use client-side printing.' % self.session_info.username)

        spool_dir = os.path.join(self.session_info.local_container, 'spool')
        if not os.path.exists(spool_dir):
            os.makedirs(spool_dir)
        self.share_local_folder(local_path=spool_dir, folder_type='spool')
        self.print_queue = printqueue.X2GoPrintQueue(profile_name=self.profile_name,
                                                     session_name=self.session_info.name,
                                                     spool_dir=spool_dir,
                                                     print_action=self.print_action, 
                                                     print_action_args=self.print_action_args,
                                                     client_instance=self.client_instance,
                                                     printing_backend=self.printing_backend,
                                                     logger=self.logger,
                                                    )
        self.print_queue.start()
        self.active_threads.append(self.print_queue)

    def set_print_action(self, print_action, **kwargs):
        """\
        Set a print action for the next incoming print jobs.

        This method is a wrapper for L{X2GoPrintQueue}C{.set_print_action()}.

        @param print_action: print action name or object (i.e. an instance of C{X2GoPrintAction*} classes)
        @type print_action: C{str} or C{X2GoPrintAction*}
        @param kwargs: print action specific parameters
        @type kwargs: dict

        """
        self.print_queue.set_print_action(print_action, logger=self.logger, **kwargs)

    def stop_printing(self):
        """\
        Shutdown (pause) the X2Go Print Queue thread.

        """
        if self.print_queue is not None:
            self.print_queue.pause()

    def get_printing_spooldir(self):
        """\
        Return the server-side printing spooldir path.

        @return: the directory for remote print job spooling
        @rtype: C{str}

        """
        return '%s/%s' % (self.session_info.remote_container, 'spool')

    def start_mimebox(self, mimebox_extensions=[], mimebox_action=None):
        """\
        Initialize the X2Go MIME box. Open/process incoming files from the server-side locally.

        @param mimebox_extensions: file name extensions that are allowed for local opening/processing
        @type mimebox_extensions: C{list}
        @param mimebox_action: MIME box action given as name or object (i.e. an instance of C{X2GoMIMEboxAction*} classes).
        @type mimebox_action: C{str} or C{obj}

        @raise X2GoUserException: if the X2Go MIME box feature is not available to this user

        """
        if not self.control_session.is_sshfs_available():
            raise x2go_exceptions.X2GoUserException('Remote user %s is not allowed to use the MIME box.' % self.session_info.username)

        mimebox_dir = os.path.join(self.session_info.local_container, 'mimebox')
        if not os.path.exists(mimebox_dir):
            os.makedirs(mimebox_dir)
        self.share_local_folder(local_path=mimebox_dir, folder_type='mimebox')
        self.mimebox_queue = mimebox.X2GoMIMEboxQueue(profile_name=self.profile_name,
                                                      session_name=self.session_info.name,
                                                      mimebox_dir=mimebox_dir,
                                                      mimebox_extensions=mimebox_extensions,
                                                      mimebox_action=mimebox_action,
                                                      client_instance=self.client_instance,
                                                      logger=self.logger,
                                                     )
        self.mimebox_queue.start()
        self.active_threads.append(self.mimebox_queue)

    def set_mimebox_action(self, mimebox_action, **kwargs):
        """\
        Set a MIME box action for the next incoming MIME jobs.

        This method is a wrapper for L{X2GoMIMEboxQueue}C{set_mimebox_action()}.

        @param mimebox_action: MIME box action name or object (i.e. an instance of C{X2GoMIMEboxAction*} classes)
        @type mimebox_action: C{str} or C{X2GoMIMEboxAction*}
        @param kwargs: MIME box action specific parameters
        @type kwargs: dict

        """
        self.mimebox_queue.set_mimebox_action(mimebox_action, logger=self.logger, **kwargs)

    def stop_mimebox(self):
        """\
        Shutdown (pause) the X2Go MIME box Queue thread.

        """
        if self.mimebox_queue is not None:
            self.mimebox_queue.pause()

    def get_mimebox_spooldir(self):
        """\
        Return the server-side MIME box spooldir path.

        @return: the directory where remote MIME box jobs are placed
        @rtype: C{str}

        """
        return '%s/%s' % (self.session_info.remote_container, 'mimebox')

    def is_session_info_protected(self):
        """\
        Test if this terminal's session info object is write-protected.

        @return: C{True}, if session info object is read-only, C{False} for read-write.
        @rtype: C{bool}

        """
        self.session_info.is_protected()

    def session_info_protect(self):
        """\
        Protect this terminal session's info object against updates.

        """
        self.session_info.protect()

    def session_info_unprotect(self):
        """\
        Allow session info updates from within the list_sessions method of the control session.

        """
        self.session_info.unprotect()

    def share_local_folder(self, local_path=None, folder_type='disk'):
        """\
        Share a local folder with the X2Go session.

        @param local_path: the full path to an existing folder on the local 
            file system
        @type local_path: C{str}
        @param folder_type: one of 'disk' (a folder on your local hard drive), 'rm' (removeable device), 
            'cdrom' (CD/DVD Rom) or 'spool' (for X2Go print spooling)
        @type folder_type: C{str}

        @return: returns C{True} if the local folder has been successfully mounted within the X2Go server session
        @rtype: C{bool}

        @raise X2GoUserException: if local folder sharing is not available to this user
        @raise Exception: any other exception occuring on the way is passed through by this method

        """
        if not self.control_session.is_sshfs_available():
            raise x2go_exceptions.X2GoUserException('Remote user %s is not allowed to share local folders with the server.' % self.session_info.username)

        if local_path is None:
            self.logger('no folder name given...', log.loglevel_WARN)
            return False

        if type(local_path) not in (types.StringType, types.UnicodeType):
            self.logger('folder name needs to be of type StringType...', log.loglevel_WARN)
            return False

        if not os.path.exists(local_path):
            self.logger('local folder does not exist: %s' % local_path, log.loglevel_WARN)
            return False

        local_path = os.path.normpath(local_path)
        self.logger('sharing local folder: %s' % local_path, log.loglevel_INFO)

        _auth_rsakey = self.control_session._x2go_session_auth_rsakey
        _host_rsakey = defaults.RSAHostKey

        _tmp_io_object = cStringIO.StringIO()
        _auth_rsakey.write_private_key(_tmp_io_object)
        _tmp_io_object.write('----BEGIN RSA IDENTITY----')
        _tmp_io_object.write('%s %s' % (_host_rsakey.get_name(),_host_rsakey.get_base64(),))

        # _x2go_key_fname must be a UniX path
        _x2go_key_fname = '%s/%s/%s' % (os.path.dirname(self.session_info.remote_container), 'ssh', 'key.z%s' % self.session_info.agent_pid)
        _x2go_key_bundle = _tmp_io_object.getvalue()

        # if there is another call to this method currently being processed, wait for that one to finish
        self._share_local_folder_lock.acquire()

        try:
            self.control_session._x2go_sftp_write(_x2go_key_fname, _x2go_key_bundle)

            _convert_encoding = self.params.convert_encoding
            _client_encoding = self.params.client_encoding
            _server_encoding = self.params.server_encoding

            if _X2GOCLIENT_OS == 'Windows':
                if local_path.startswith('\\\\'):
                    # we are on a UNC path
                    if 'X2GO_MOUNT_UNCPATHS' in self.control_session.get_server_features():
                        local_path = local_path.repalce('\\\\', '/uncpath/')
                    else:
                        local_path = local_path.repalce('\\\\', '/windrive/')
                    local_path = local_path.replace('\\', '/')
                else:
                    local_path = local_path.replace('\\', '/')
                    local_path = local_path.replace(':', '')
                    local_path = '/windrive/%s' % local_path
                _convert_encoding = True
                _client_encoding = 'WINDOWS-1252'

            if _convert_encoding:
                export_iconv_settings = 'export X2GO_ICONV=modules=iconv,from_code=%s,to_code=%s && ' % (_client_encoding, _server_encoding)
            else:
                export_iconv_settings = ''

            if folder_type == 'disk':

                cmd_line = [ '%sexport HOSTNAME &&' % export_iconv_settings,
                             'x2gomountdirs',
                             'dir',
                             str(self.session_info.name),
                             '\'%s\'' % _CURRENT_LOCAL_USER,
                             _x2go_key_fname,
                             '%s__REVERSESSH_PORT__%s; ' % (local_path, self.session_info.sshfs_port),
                             'rm -f %s %s.ident' % (_x2go_key_fname, _x2go_key_fname),
                           ]

            elif folder_type == 'spool':

                cmd_line = [ '%sexport HOSTNAME &&' % export_iconv_settings,
                             'x2gomountdirs',
                             'dir',
                             str(self.session_info.name),
                             '\'%s\'' % _CURRENT_LOCAL_USER,
                             _x2go_key_fname,
                             '%s__PRINT_SPOOL___REVERSESSH_PORT__%s; ' % (local_path, self.session_info.sshfs_port),
                             'rm -f %s %s.ident' % (_x2go_key_fname, _x2go_key_fname), 
                           ]

            elif folder_type == 'mimebox':

                cmd_line = [ '%sexport HOSTNAME &&' % export_iconv_settings,
                             'x2gomountdirs',
                             'dir',
                             str(self.session_info.name),
                             '\'%s\'' % _CURRENT_LOCAL_USER,
                             _x2go_key_fname,
                             '%s__MIMEBOX_SPOOL___REVERSESSH_PORT__%s; ' % (local_path, self.session_info.sshfs_port),
                             'rm -f %s %s.ident' % (_x2go_key_fname, _x2go_key_fname), 
                           ]

            (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)
            _stdout = stdout.read().split('\n')
            self.logger('x2gomountdirs output is: %s' % _stdout, log.loglevel_NOTICE)

        except:
            self._share_local_folder_lock.release()
            raise
        self._share_local_folder_lock.release()

        if len(_stdout) >= 6 and _stdout[5].endswith('ok'):
            return True
        return False

    def unshare_all_local_folders(self):
        """\
        Unshare all local folders mount in the X2Go session.

        @return: returns C{True} if all local folders could be successfully unmounted from the X2Go server session
        @rtype: C{bool}

        """
        self.logger('unsharing all local folders from session %s' % self.session_info, log.loglevel_INFO)

        cmd_line = [ 'export HOSTNAME &&',
                     'x2goumount-session', 
                     self.session_info.name,
                   ]

        (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)
        if not stderr.read():
            self.logger('x2goumount-session (all mounts) for session %s has been successful' % self.session_info, log.loglevel_NOTICE)
            return True
        else:
            self.logger('x2goumount-session (all mounts) for session %s failed' % self.session_info, log.loglevel_ERROR)
            return False

    def unshare_local_folder(self, local_path):
        """\
        Unshare local folder given as <local_path> from X2Go session.

        @return: returns C{True} if the local folder <local_path> could be successfully unmounted from the X2Go server session
        @rtype: C{bool}

        """
        self.logger('unsharing local folder from session %s' % self.session_info, log.loglevel_INFO)

        cmd_line = [ 'export HOSTNAME &&',
                     'x2goumount-session', 
                     self.session_info.name,
                     "'%s'" % local_path,
                   ]

        (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)
        if not stderr.read():
            self.logger('x2goumount-session (%s) for session %s has been successful' % (local_path, self.session_info, ), log.loglevel_NOTICE)
            return True
        else:
            self.logger('x2goumount-session (%s) for session %s failed' % (local_path, self.session_info, ), log.loglevel_ERROR)
            return False

    def color_depth(self):
        """\
        Retrieve the session's color depth.

        @return: the session's color depth
        @rtype: C{int}

        """
        return self.params.depth

    def auto_session_window_title(self, dont_set=False):
        """\
        Automatically generate an appropriate human-readable session window title.

        The session window title will be provider in the C{session_title} property of
        this method.

        @param dont_set: generate the session window title, but do not actually set it
        @type dont_set: C{bool}

        """
        _generic_title = 'X2GO-%s' % self.session_info.name

        # no blanks at beginning or end, no blanks-only...
        self.session_title = self.session_title.strip()

        if self.params.session_type == 'D':
            if self.set_session_title:

                if not self.session_title:
                    self.session_title = '%s for %s@%s' % (self.params.cmd, self.control_session.remote_username(), self.control_session.get_hostname())

            else:
                # session title fallback... (like X2Go server does it...)
                self.session_title = _generic_title

        elif self.params.session_type == 'S':
            if self.set_session_title:

                shared_user = _generic_title.split('XSHAD')[1]
                shared_display = _generic_title.split('XSHAD')[2].replace('PP', ':').split("_")[0]

                self.session_title = 'Desktop %s@%s shared with %s@%s' % (shared_user, shared_display, self.control_session.remote_username(), self.control_session.get_hostname())

            else:
                # session title fallback... (like X2Go server does it...)
                self.session_title = _generic_title

        else:
            # do nothing for rootless sessions
            self.session_title = _generic_title

        if self.session_title != _generic_title and not dont_set:
            self.set_session_window_title(title=self.session_title)

    def find_session_window(self, timeout=60):
        """\
        Try for <timeout> seconds to find the X2Go session window of this
        terminal session.

        A background thread will get spawned for this operation.

        @param timeout: try for <timeout> seconds to find the session window
        @type timeout: C{int}

        """
        gevent.spawn(self._find_session_window, timeout=timeout)

    def _find_session_window(self, timeout=0):
        """\
        Try for <timeout> seconds to find the X2Go session window of this
        terminal session.

        @param timeout: try for <timeout> seconds to find the session window
        @type timeout: C{int}

        """
        self.session_window = None

        # search for the window of our focus, do this in a loop till the window as been found
        # or timeout forces us to give up...
        timeout += 1
        while timeout:

            timeout -= 1

            window = utils.find_session_window(self.session_info.name)

            if window is not None:
                self.session_window = window
                break

            gevent.sleep(1)

    def set_session_window_title(self, title, timeout=60):
        """\
        Modify the session window title.

        A background thread will get spawned for this operation.

        @param title: new title for the terminal session's session window
        @type title: C{str}
        @param timeout: try for <timeout> seconds to find the session window
        @type timeout: C{int}

        """
        gevent.spawn(self._set_session_window_title, title=title.strip(), timeout=timeout)

    def _set_session_window_title(self, title, timeout=0):
        """\
        Modify the session window title.

        @param title: new title for the terminal session's session window
        @type title: C{str}
        @param timeout: try for <timeout> seconds to find the session window
        @type timeout: C{int}

        """
        self.session_title = title

        if not self.session_title:
            self.auto_session_title(dont_set=True)

        timeout += 1
        while timeout:

            timeout -= 1

            if self.session_window is not None:
                utils.set_session_window_title(self.session_window, self.session_title)
                break

            gevent.sleep(1)

    def raise_session_window(self, timeout=60):
        """\
        Try for <timeout> seconds to raise the X2Go session window of this
        terminal session to the top and bring it to focus.

        A background thread will get spawned for this operation.

        @param timeout: try for <timeout> seconds to raise the session window
        @type timeout: C{int}

        """
        gevent.spawn(self._raise_session_window, timeout=timeout)

    def _raise_session_window(self, timeout=0):
        """
        Try for <timeout> seconds to raise the X2Go session window of this
        terminal session to the top and bring it to focus.

        @param timeout: try for <timeout> seconds to raise the session window
        @type timeout: C{int}

        """
        timeout += 1
        while timeout:

            timeout -= 1

            if self.session_window is not None:

                utils.raise_session_window(self.session_window)
                break

            gevent.sleep(1)

    def has_command(self, cmd):
        """\
        ,,Guess'' if the command C{<cmd>} exists on the X2Go server and is executable.
        The expected result is not 100% safe, however, it comes with a high probability to
        be correct.

        @param cmd: session command
        @type cmd: C{str}

        @return: C{True} if this method reckons that the command is executable on the remote X2Go server
        @rtype: C{bool}

        """
        test_cmd = None;

        cmd = cmd.strip('"').strip('"')
        if cmd.find('RDP') != -1:
            cmd = 'rdesktop'

        if cmd in _X2GO_GENERIC_APPLICATIONS:
            return True
        if cmd in _X2GO_DESKTOPSESSIONS.keys():
            return True
        elif 'XSHAD' in cmd:
            return True
        elif 'PUBLISHED' in cmd and 'X2GO_PUBLISHED_APPLICATIONS' in self.control_session.get_server_features():
            return True
        elif cmd and cmd.startswith('/'):
            # check if full path is correct _and_ if application is in server path
            test_cmd = 'test -x %s && which %s && echo OK' % (cmd, os.path.basename(cmd.split()[0]))
        elif cmd and '/' not in cmd.split()[0]:
            # check if application is in server path only
            test_cmd = 'which %s && echo OK' % os.path.basename(cmd.split()[0])

        if test_cmd:
            (stdin, stdout, stderr) = self.control_session._x2go_exec_command([test_cmd])
            _stdout = stdout.read()
            return _stdout.find('OK') != -1
        else:
            return False

    def run_command(self, cmd=None, env={}):
        """\
        Run a command in this session.

        After L{X2GoTerminalSessionSTDOUT.start()} has been called 
        one or more commands can be executed with L{X2GoTerminalSessionSTDOUT.run_command()}
        within the current X2Go session.

        @param cmd: Command to be run
        @type cmd: C{str}
        @param env: add server-side environment variables
        @type env: C{dict}

        @return: stdout.read() and stderr.read() as returned by the run command
            on the X2Go server
        @rtype: C{tuple} of C{str}

        """
        if not self.has_command(_rewrite_cmd(str(self.params.cmd), params=self.params)):
            if self.client_instance:
                self.client_instance.HOOK_no_such_command(profile_name=self.profile_name, session_name=self.session_info.name, cmd=self.params.cmd)
            return False

        if cmd in ("", None):
            if self.params.cmd is None:
                cmd = 'TERMINAL'
            else:
                cmd = self.params.cmd

        if cmd == 'XDMCP':
            # do not run command when in XDMCP mode...
            return None

        if 'XSHAD' in cmd:
            # do not run command when in DESKTOP SHARING mode...
            return None

        self.params.update(cmd=cmd)

        # do not allow the execution of full path names
        if '/' in cmd:
            cmd = os.path.basename(cmd)

        cmd_line = [ "setsid x2goruncommand", 
                     str(self.session_info.display),
                     str(self.session_info.agent_pid),
                     str(self.session_info.name), 
                     str(self.session_info.snd_port),
                     _rewrite_blanks(_rewrite_cmd(cmd, params=self.params)),
                     str(self.params.snd_system),
                     str(self.params.session_type),
                     "1>/dev/null 2>/dev/null & exit",
                   ]

        if self.params.snd_system == 'pulse':
            cmd_line = [ 'PULSE_CLIENTCONFIG=%s/.pulse-client.conf' % self.session_info.remote_container ] + cmd_line

        if env:
            for env_var in env.keys():
                cmd_line = [ '%s=%s' % (env_var, env[env_var]) ] + cmd_line

        (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)

        if self.params.kbtype not in ('null/null', 'auto') and (self.params.kblayout not in ('null', '') or self.params.kbvariant not in ('null', '')):
            self.set_keyboard(layout=self.params.kblayout, variant=self.params.kbvariant)

        return stdout.read(), stderr.read()

    def is_desktop_session(self):
        """\
        Is this (terminal) session a desktop session?

        @return: Returns C{True} is this session is a desktop session.
        @rtype: C{bool}

        """
        if self.session_info:
            return self.session_info.is_desktop_session()
        return False

    def is_published_applications_provider(self):
        """\
        Is this (terminal) session a published applications provider?

        @return: Returns C{True} is this session is a provider session for published applications.
        @rtype: C{bool}

        """
        if self.session_info and self.is_running():
            return self.session_info.is_published_applications_provider()
        return False

    def set_keyboard(self, layout='null', variant='null'):
        """\
        Set the keyboard layout and variant for this (running) session.

        @param layout: keyboard layout to be set
        @type layout: C{str}
        @param variant: keyboard variant to be set
        @type variant: C{str}

        @return: returns C{True} if the {setxkbmap} command could be executed successfully.
        @rtype: C{bool}

        """
        if not self.is_running():
            return False

        cmd_line = [ 'export DISPLAY=:%s && ' % str(self.session_info.display),
                     'setxkbmap '
                   ]

        if layout != 'null':
            self.logger('setting keyboad layout ,,%s\'\' for session %s' % (layout, self.session_info), log.loglevel_INFO)
            cmd_line.append('-layout %s' % layout)
        if variant != 'null':
            self.logger('setting keyboad variant ,,%s\'\' for session %s' % (variant, self.session_info), log.loglevel_INFO)
            cmd_line.append('-variant %s' % variant)

        (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)
        _stderr = stderr.read()
        if not _stderr:
            self.logger('setting keyboard layout ,,%s\'\' and variant ,,%s\'\' for session %s has been successful' % (layout, variant, self.session_info), log.loglevel_NOTICE)
            return True
        else:
            self.logger('setting keyboard layout ,,%s\'\' and variant ,,%s\'\' for session %s failed: %s' % (layout, variant, self.session_info, _stderr.replace('\n', ' ')), log.loglevel_ERROR)
            return False

    def exec_published_application(self, exec_name, timeout=20, env={}):
        """\
        Executed a published application.

        @param exec_name: application to be executed
        @type exec_name: C{str}
        @param timeout: execution timeout
        @type timeout: C{int}
        @param env: session environment dictionary
        @type env: C{dict}

        """
        cmd_line = [
            "export DISPLAY=:%s && " % str(self.session_info.display),
            "export X2GO_SESSION=%s && " % str(self.get_session_name()),
        ]

        if self.params.snd_system == 'pulse':
            cmd_line.append("export PULSE_CLIENTCONFIG=%s/.pulse-client.conf && " % self.session_info.remote_container)

        if env:
            for env_var in env.keys():
                cmd_line = [ 'export %s=%s && ' % (env_var, env[env_var]) ] + cmd_line

        cmd_line.extend(
            [
                "setsid %s" % exec_name,
                "1>/dev/null 2>/dev/null & exit",
            ]
        )

        self.logger('executing published application %s for %s with command line: %s' % (exec_name, self.profile_name, cmd_line), loglevel=log.loglevel_DEBUG)
        (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line, timeout=timeout)

    def ok(self):
        """\
        X2Go session OK?

        @return: Returns C{True} if this X2Go (terminal) session is up and running,
            C{False} otherwise.
        @rtype: C{bool}

        """
        _ok = bool(self.session_info.name and self.proxy.ok())
        return _ok

    def is_running(self):
        """\
        X2Go session running?

        @return: Returns C{True} if this X2Go (terminal) session is in running state,
            C{False} otherwise.
        @rtype: C{bool}

        """
        return self.session_info.is_running()

    def is_suspended(self):
        """\
        X2Go session suspended?

        @return: Returns C{True} if this X2Go (terminal) session is in suspended state,
            C{False} otherwise.
        @rtype: C{bool}

        """
        return self.session_info.is_suspended()

    def is_connected(self):
        """\
        X2Go session connected?

        @return: Returns C{True} if this X2Go session's Paramiko/SSH transport is 
            connected/authenticated, C{False} else.
        @rtype: C{bool}

        """
        return self.control_session.is_connected()

    def start(self):
        """\
        Start a new X2Go session.

        @return: C{True} if session startup has been successful and the X2Go proxy is up-and-running
        @rtype: C{bool}

        @raise X2GoTerminalSessionException: if the session startup failed

        """
        self.params.rewrite_session_type()

        if not self.has_command(_rewrite_cmd(self.params.cmd, params=self.params)):
            if self.client_instance:
                self.client_instance.HOOK_no_such_command(profile_name=self.profile_name, session_name=self.session_info.name, cmd=self.params.cmd)
            return False

        setkbd = "0"
        if self.params.kbtype != "null/null":
            setkbd = "1"

        if '/' in self.params.cmd:
            self.params.cmd = os.path.basename(self.params.cmd)

        self.params.rewrite_session_type()

        if self.params.geometry == 'maximize':
            _geometry = utils.get_workarea_geometry()
            if _geometry is None or len(_geometry) != 2:
                _geometry = utils.get_desktop_geometry()
            if _geometry and len(_geometry) == 2:
                self.params.geometry = "%sx%s" % _geometry
            else:
                self.logger('failed to detect best maxmimized geometry of your client-side desktop', loglevel=log.loglevel_WARN)
                self.params.geometry = "1024x768"

        cmd_line = [ "x2gostartagent",
                     str(self.params.geometry),
                     str(self.params.link),
                     str(self.params.pack),
                     str(self.params.cache_type+'-depth_'+self.params.depth),
                     str(self.params.kblayout),
                     str(self.params.kbtype),
                     str(setkbd),
                     str(self.params.session_type),
                     str(self.params.cmd),
                   ]

        if self.params.cmd == 'XDMCP' and self.params.xdmcp_server:
            cmd_line = ['X2GOXDMCP=%s' % self.params.xdmcp_server] + cmd_line

        if self.params.dpi:
            cmd_line = ['X2GODPI=%s' % self.params.dpi] + cmd_line

        (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)

        _stdout = stdout.read()
        _stderr = stderr.read()

        # if the first line of stdout is a "DEN(Y)" string then we will presume that
        # we tried to use X2Go desktop sharing and the sharing was rejected
        if "ACCESS DENIED" in _stderr and "XSHAD" in _stderr:
            raise x2go_exceptions.X2GoDesktopSharingException('X2Go desktop sharing has been denied by the remote user')

        try:
            self.session_info.initialize(_stdout,
                                         username=self.control_session.remote_username(),
                                         hostname=self.control_session.remote_peername(),
                                        )
        except ValueError:
            raise x2go_exceptions.X2GoTerminalSessionException("failed to start X2Go session")
        except IndexError:
            raise x2go_exceptions.X2GoTerminalSessionException("failed to start X2Go session")

        # local path may be a Windows path, so we use the path separator of the local system
        self.session_info.local_container = os.path.join(self.params.rootdir, 'S-%s' % self.session_info.name)
        # remote path is always a UniX path...
        self.session_info.remote_container = '%s/.x2go/C-%s' % (self.control_session._x2go_remote_home,
                                                                self.session_info.name,
                                                               )

        # set up SSH tunnel for X11 graphical elements
        self.proxy = self.proxy_backend(session_info=self.session_info, 
                                        ssh_transport=self.control_session.get_transport(),
                                        sessions_rootdir=self.sessions_rootdir,
                                        session_instance=self.session_instance,
                                        proxy_options=self.proxy_options,
                                        logger=self.logger)
        self.proxy_subprocess, proxy_ok = self.proxy.start_proxy()

        if proxy_ok:
            self.active_threads.append(self.proxy)

            if self.params.session_type in ('D', 'S'):
                self.find_session_window()
                self.auto_session_window_title()
                self.raise_session_window()

            if self.params.published_applications:
                self.control_session.get_published_applications()

        else:
            raise x2go_exceptions.X2GoTerminalSessionException("failed to start X2Go session")

        return proxy_ok

    def resume(self):
        """\
        Resume a running/suspended X2Go session. 

        @return: C{True} if the session could successfully be resumed
        @rtype: C{bool}

        @raise X2GoTerminalSessionException: if the terminal session failed to update server-side reported port changes

        """
        setkbd = "0"
        if self.params.kbtype != "null/null":
            setkbd = "1"

        if self.params.geometry == 'maximize':
            _geometry = utils.get_workarea_geometry()
            if _geometry is None or len(_geometry) != 2:
                _geometry = utils.get_desktop_geometry()
            if _geometry and len(_geometry) == 2:
                self.params.geometry = "%sx%s" % _geometry
            else:
                self.logger('failed to detect best maxmimized geometry of your client-side desktop, using 1024x768 instead', loglevel=log.loglevel_WARN)
                self.params.geometry = "1024x768"

        cmd_line = [ "x2goresume-session", self.session_info.name,
                     self.params.geometry,
                     self.params.link,
                     self.params.pack,
                     self.params.kblayout,
                     self.params.kbtype,
                     setkbd,
                   ]

        (stdin, stdout, stderr) = self.control_session._x2go_exec_command(cmd_line)

        # re-allocate (if needed) server-side ports for graphics, sound and sshfs
        for stdout_line in stdout.read():
            try:
                _new_value = stdout_line.split("=")[1].strip()
                if 'gr_port=' in stdout_line and _new_value != str(self.session_info.graphics_port):
                    try:
                        self.session_info.graphics_port = int(_new_value)
                        self.logger('re-allocating graphics port for session %s, old server-side port is in use; new graphics port is %s' % (self.session_info, self.session_info.graphics_port), loglevel=log.loglevel_NOTICE)
                    except TypeError:
                        # if the re-allocation fails, this is fatal!!!
                        raise x2go_exceptions.X2GoTerminalSessionException('Failed to retrieve new graphics port from server. X2Go Session cannot be resumed.')
                elif 'sound_port=' in stdout_line and _new_value != str(self.session_info.snd_port):
                    try:
                        self.session_info.snd_port = int(_new_value)
                        self.logger('re-allocating sound port for session %s, old server-side port is in use; new sound port is %s' % (self.session_info, self.session_info.snd_port), loglevel=log.loglevel_NOTICE)
                    except TypeError:
                        self.logger('Failed to retrieve new sound port from server for session %s, session will be without sound.' % self.session_info, loglevel=log.loglevel_WARN)
                elif 'fs_port=' in stdout_line and _new_value != str(self.session_info.sshfs_port):
                    try:
                        self.session_info.sshfs_port = int(_new_value)
                        self.logger('re-allocating sshfs port for session %s, old server-side port is in use; new sshfs port is %s' % (self.session_info, self.session_info.sshfs_port), loglevel=log.loglevel_NOTICE)
                    except TypeError:
                        self.logger('Failed to retrieve new sshfs port from server for session %s, session will be without client-side folder sharing. Neither will there be X2Go printing nor X2Go MIME box support.' % self.session_info, loglevel=log.loglevel_WARN)
            except IndexError:
                continue

        # local path may be a Windows path, so we use the path separator of the local system
        self.session_info.local_container = os.path.join(self.params.rootdir, 'S-%s' % self.session_info.name)
        # remote path is always a UniX path...
        self.session_info.remote_container = '%s/.x2go/C-%s' % (self.control_session._x2go_remote_home, 
                                                                self.session_info.name,
                                                               )
        self.proxy = self.proxy_backend(session_info=self.session_info, 
                                        ssh_transport=self.control_session.get_transport(), 
                                        sessions_rootdir=self.sessions_rootdir,
                                        session_instance=self.session_instance,
                                        logger=self.logger
                                       )
        self.proxy_subprocess, proxy_ok = self.proxy.start_proxy()

        if proxy_ok:
            self.params.depth = self.session_info.name.split('_')[2][2:]

            # on a session resume the user name comes in as a user ID. We have to translate this...
            self.session_info.username = self.control_session.remote_username()

            if self.params.kbtype not in ('null/null', 'auto') and (self.params.kblayout not in ('null', '') or self.params.kbvariant not in ('null', '')):
                self.set_keyboard(layout=self.params.kblayout, variant=self.params.kbvariant)

            if self.params.session_type in ('D', 'S'):
                self.find_session_window()
                self.auto_session_window_title()
                self.raise_session_window()

            if self.is_published_applications_provider():
                self.control_session.get_published_applications()
                self.published_applications = True
        else:
            raise x2go_exceptions.X2GoTerminalSessionException("failed to start X2Go session")

        return proxy_ok

    def suspend(self):
        """\
        Suspend this X2Go (terminal) session.

        @return: C{True} if the session terminal could be successfully suspended
        @rtype: C{bool}

        """
        self.control_session.suspend(session_name=self.session_info.name)
        self.release_proxy()

        # TODO: check if session has really suspended
        _ret = True

        return _ret

    def terminate(self):
        """\
        Terminate this X2Go (terminal) session.

        @return: C{True} if the session could be successfully terminated
        @rtype: C{bool}

        """
        self.control_session.terminate(session_name=self.session_info.name, destroy_terminals=False)
        self.release_proxy()
        self.post_terminate_cleanup()
        self.__del__()

        # TODO: check if session has really suspended
        _ret = True

        return _ret

    def release_proxy(self):
        """\
        Let the X2Go proxy command cleanly die away... (by calling its destructor).

        """
        if self.proxy is not None:
            self.proxy.__del__()

    def post_terminate_cleanup(self):
        """\
        Do some cleanup after this session has terminated.

        """
        # this method might be called twice (directly and from update_status in the session
        # registry instance. So we have to make sure, that this code will not fail
        # if called twice.
        if not self._cleaned_up and self.session_info.name:

            # otherwise we wipe the session files locally
            self.logger('cleaning up session %s after termination' % self.session_info, loglevel=log.loglevel_NOTICE)

            # if we run in debug mode, we keep local session directories
            if self.logger.get_loglevel() & log.loglevel_DEBUG != log.loglevel_DEBUG:

                self._rm_session_dirtree()
                self._rm_desktop_dirtree()

            self._cleaned_up = True

    def is_rootless_session(self):
        """\
        Test if this terminal session is a rootless session.

        @return: C{True} if this session is of session type rootless ('R').
        @rtype: C{bool}

        """
        self.params.rewrite_session_type()
        return self.params.session_type == 'R'

    def is_shadow_session(self):
        """\
        Test if this terminal session is a desktop sharing (aka shadow) session.

        @return: C{True} if this session is of session type shadow ('S').
        @rtype: C{bool}

        """
        self.params.rewrite_session_type()
        return self.params.session_type == 'S'

    def is_pubapp_session(self):
        """\
        Test if this terminal session is a published applications session.

        @return: C{True} if this session is of session type published applications ('P').
        @rtype: C{bool}

        """
        self.params.rewrite_session_type()
        return self.params.session_type == 'P'




# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

# gevent_subprocess was found here: http://groups.google.com/group/gevent/browse_thread/thread/dba1a5d29e0a60ff
# Mark Visser <mjmvisser@gmail.com>

###
### Mark Visser, Sat, 20 Nov 2010 13:30:16 -0500
###
### <quote>
### Hi Mike,
###
### I hereby place that code snippet in the public domain, feel free to apply any license that is appropriate!
###
### cheers,
### -Mark
### </quote>
###
### Thus, I place myself as the copyright holder for code in this file
### for cases in that it is used in context of the X2Go project.
###


"""Implementation of the standard :mod:`subprocess` module that spawns greenlets"""
import errno
import sys
import fcntl, os

_subprocess = __import__('subprocess')

from gevent import socket, select, hub

# identical to original
CalledProcessError = _subprocess.CalledProcessError
MAXFD = _subprocess.MAXFD
PIPE = _subprocess.PIPE
STDOUT = _subprocess.STDOUT
call = _subprocess.call
check_call = _subprocess.check_call
list2cmdline = _subprocess.list2cmdline


class Popen(object):
    def __init__(self, *args, **kwargs):
        # delegate to an actual Popen object
        self.__p = _subprocess.Popen(*args, **kwargs)
        # make the file handles nonblocking
        if self.stdin is not None:
            fcntl.fcntl(self.stdin, fcntl.F_SETFL, os.O_NONBLOCK)
        if self.stdout is not None:
            fcntl.fcntl(self.stdout, fcntl.F_SETFL, os.O_NONBLOCK)
        if self.stderr is not None:
            fcntl.fcntl(self.stderr, fcntl.F_SETFL, os.O_NONBLOCK)

    def __getattr__(self, name):
        # delegate attribute lookup to the real Popen object
        return getattr(self.__p, name)

    def _write_pipe(self, f, input):
        # writes the given input to f without blocking
        if input:
            bytes_total = len(input)
            bytes_written = 0
            while bytes_written < bytes_total:
                try:
                    # f.write() doesn't return anything, so use os.write.
                    bytes_written += os.write(f.fileno(), input[bytes_written:])
                except IOError, ex:
                    if ex[0] != errno.EAGAIN:
                        raise
                    sys.exc_clear()
                socket.wait_write(f.fileno())
        f.close()

    def _read_pipe(self, f):
        # reads output from f without blocking
        # returns output
        chunks = []
        while True:
            try:
                chunk = f.read(4096)
                if not chunk:
                    break
                chunks.append(chunk)
            except IOError, ex:
                if ex[0] != errno.EAGAIN:
                    raise
                sys.exc_clear()
            socket.wait_read(f.fileno())
        f.close()
        return ''.join(chunks)

    def communicate(self, input=None):
        # Optimization: If we are only using one pipe, or no pipe at
        # all, using select() is unnecessary.
        if [self.stdin, self.stdout, self.stderr].count(None) >= 2:
            stdout = None
            stderr = None
            if self.stdin:
                self._write_pipe(self.stdin, input)
            elif self.stdout:
                stdout = self._read_pipe(self.stdout)
            elif self.stderr:
                stderr = self._read_pipe(self.stderr)
            self.wait()
            return (stdout, stderr)
        else:
            return self._communicate(input)

    def _communicate(self, input):
        # identical to original... all the heavy lifting is done
        # in gevent.select.select
        read_set = []
        write_set = []
        stdout = None # Return
        stderr = None # Return

        if self.stdin:
            # Flush stdin buffer.
            self.stdin.flush()
            if input:
                write_set.append(self.stdin)
            else:
                self.stdin.close()
        if self.stdout:
            read_set.append(self.stdout)
            stdout = []
        if self.stderr:
            read_set.append(self.stderr)
            stderr = []

        input_offset = 0
        while read_set or write_set:
            try:
                rlist, wlist, xlist = select.select(read_set, write_set, [])
            except select.error, e:
                if e.args[0] == errno.EINTR:
                    continue
                raise

            if self.stdin in wlist:
                # When select has indicated that the file is writable,
                # we can write up to PIPE_BUF bytes without risk
                # blocking.  POSIX defines PIPE_BUF >= 512
                bytes_written = os.write(self.stdin.fileno(), buffer(input, input_offset, 512))
                input_offset += bytes_written
                if input_offset >= len(input):
                    self.stdin.close()
                    write_set.remove(self.stdin)

            if self.stdout in rlist:
                data = os.read(self.stdout.fileno(), 1024)
                if data == "":
                    self.stdout.close()
                    read_set.remove(self.stdout)
                stdout.append(data)

            if self.stderr in rlist:
                data = os.read(self.stderr.fileno(), 1024)
                if data == "":
                    self.stderr.close()
                    read_set.remove(self.stderr)
                stderr.append(data)

        # All data exchanged.  Translate lists into strings.
        if stdout is not None:
            stdout = ''.join(stdout)
        if stderr is not None:
            stderr = ''.join(stderr)

        # Translate newlines, if requested.  We cannot let the file
        # object do the translation: It is based on stdio, which is
        # impossible to combine with select (unless forcing no
        # buffering).
        if self.universal_newlines and hasattr(file, 'newlines'):
            if stdout:
                stdout = self._translate_newlines(stdout)
            if stderr:
                stderr = self._translate_newlines(stderr)

        self.wait()
        return (stdout, stderr)

    def wait(self, check_interval=0.01):
        # non-blocking, use hub.sleep
        try:
            while True:
                status = self.poll()
                if status >= 0:
                    return status
                hub.sleep(check_interval)
        except OSError, e:
            if e.errno == errno.ECHILD:
                # no child process, this happens if the child process
                # already died and has been cleaned up
                return -1
            else:
                raise


# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
Default variables and values for Python X2Go.

"""
__NAME__ = 'x2godefaults-pylib'

import os
import paramiko
import platform

##
## Common X2Go defaults
##

X2GOCLIENT_OS = platform.system()

if X2GOCLIENT_OS != 'Windows':
    import Xlib.display
    import Xlib.error

    # handle missing X displays on package build
    try:
        X_DISPLAY = Xlib.display.Display()
    except Xlib.error.DisplayNameError:
        X_DISPLAY = None
    except Xlib.error.DisplayConnectionError:
        X_DISPLAY = None

LOCAL_HOME = os.path.normpath(os.path.expanduser('~'))
X2GO_SESSIONS_ROOTDIR = '.x2go'
X2GO_CLIENT_ROOTDIR = '.x2goclient'
X2GO_SSH_ROOTDIR = os.path.join('.x2go','.ssh')

# setting OS dependent variables
if X2GOCLIENT_OS == "Windows":
    # on Windows we will use the current directory as ,,ROOTDIR'' which 
    # will normally be the application directory
    ROOT_DIR = os.path.abspath(os.path.curdir)
    ETC_DIR = os.path.join(ROOT_DIR, 'etc')
    import win32api
    CURRENT_LOCAL_USER = win32api.GetUserName()
    X2GO_SSH_ROOTDIR = '.ssh'
    SUPPORTED_SOUND = True
    SUPPORTED_PRINTING = True
    SUPPORTED_FOLDERSHARING = True
    SUPPORTED_MIMEBOX = True

elif X2GOCLIENT_OS == "Linux":
    ROOT_DIR = '/'
    ETC_DIR = os.path.join(ROOT_DIR, 'etc', 'x2goclient')
    import getpass
    CURRENT_LOCAL_USER = getpass.getuser()
    X2GO_SSH_ROOTDIR = '.ssh'
    SUPPORTED_SOUND = True
    SUPPORTED_PRINTING = True
    SUPPORTED_FOLDERSHARING = True
    SUPPORTED_MIMEBOX = True

elif X2GOCLIENT_OS == "Mac":
    ROOT_DIR = '/'
    ETC_DIR = os.path.join(ROOT_DIR, 'etc', 'x2goclient')
    import getpass
    CURRENT_LOCAL_USER = getpass.getuser()
    X2GO_SSH_ROOTDIR = '.ssh'
    SUPPORTED_SOUND = True
    SUPPORTED_PRINTING = True
    SUPPORTED_FOLDERSHARING = True
    SUPPORTED_MIMEBOX = True

else:
    import exceptions
    class OSNotSupportedException(exceptions.StandardError): pass
    raise OSNotSupportedException('Platform %s is not supported' % platform.system())

##
## control and terminal session backend as well as session info and proxy backend defaults
##

BACKENDS_CONTROLSESSION = {
    'STDOUT': 'X2GoControlSessionSTDOUT',
}
BACKENDS_TERMINALSESSION = {
    'STDOUT': 'X2GoTerminalSessionSTDOUT',
}
BACKENDS_SERVERSESSIONINFO = {
    'STDOUT': 'X2GoServerSessionInfoSTDOUT',
}
BACKENDS_SERVERSESSIONLIST = {
    'STDOUT': 'X2GoServerSessionListSTDOUT',
}
BACKENDS_PROXY = {
    'NX3': 'X2GoProxyNX3',
}

BACKEND_CONTROLSESSION_DEFAULT =  'X2GoControlSessionSTDOUT'
BACKEND_TERMINALSESSION_DEFAULT = 'X2GoTerminalSessionSTDOUT'
BACKEND_SERVERSESSIONINFO_DEFAULT = 'X2GoServerSessionInfoSTDOUT'
BACKEND_SERVERSESSIONLIST_DEFAULT = 'X2GoServerSessionListSTDOUT'
BACKEND_PROXY_DEFAULT = 'X2GoProxyNX3'

##
## profile backend defaults
##

BACKENDS_SESSIONPROFILES = {
    'FILE': 'X2GoSessionProfilesFILE',
    'GCONF': 'X2GoSessionProfilesGCONF',
    'HTTPSBROKER': 'X2GoSessionProfilesHTTPSBROKER',
    'WINREG': 'X2GoSessionProfilesWINREG',
}
"""Python X2Go backends for storing session profiles."""
BACKENDS_CLIENTSETTINGS = {
    'FILE': 'X2GoClientSettingsFILE',
    'GCONF': 'X2GoClientSettingsGCONF',
    'HTTPSBROKER': 'X2GoClientSettingsHTTPSBROKER',
    'WINREG': 'X2GoClientSettingsWINREG',
}
"""Python X2Go backends for storing client settings."""
BACKENDS_CLIENTPRINTING = {
    'FILE': 'X2GoClientPrintingFILE',
    'GCONF': 'X2GoClientPrintingGCONF',
    'HTTPSBROKER': 'X2GoClientPrintingHTTPSBROKER',
    'WINREG': 'X2GoClientPrintingWINREG',
}
"""Python X2Go backends for storing print settings."""

BACKEND_SESSIONPROFILES_DEFAULT = 'X2GoSessionProfilesFILE'
BACKEND_CLIENTSETTINGS_DEFAULT = 'X2GoClientSettingsFILE'
BACKEND_CLIENTPRINTING_DEFAULT = 'X2GoClientPrintingFILE'

##
## X2Go Printing
##

X2GO_SETTINGS_FILENAME = 'settings'
X2GO_SETTINGS_CONFIGFILES = [
    os.path.normpath(os.path.join(LOCAL_HOME, X2GO_CLIENT_ROOTDIR, 'settings')),
    os.path.normpath(os.path.join(ETC_DIR,X2GO_SETTINGS_FILENAME)),
]
X2GO_PRINTING_FILENAME = 'printing'
X2GO_PRINTING_CONFIGFILES = [
    os.path.normpath(os.path.join(LOCAL_HOME, X2GO_CLIENT_ROOTDIR, 'printing')),
    os.path.normpath(os.path.join(ETC_DIR,X2GO_PRINTING_FILENAME)),
]
X2GO_SESSIONPROFILES_FILENAME = 'sessions'
X2GO_SESSIONPROFILES_CONFIGFILES = [
    os.path.normpath(os.path.join(LOCAL_HOME, X2GO_CLIENT_ROOTDIR, 'sessions')),
    os.path.normpath(os.path.join(ETC_DIR,X2GO_SESSIONPROFILES_FILENAME)),
]
X2GO_XCONFIG_FILENAME = 'xconfig'
X2GO_XCONFIG_CONFIGFILES = [
    os.path.normpath(os.path.join(LOCAL_HOME, X2GO_CLIENT_ROOTDIR, 'xconfig')),
    os.path.normpath(os.path.join(ETC_DIR,X2GO_XCONFIG_FILENAME)),
]

X2GO_CLIENTSETTINGS_DEFAULTS = {
    'LDAP': {
        'useldap': False,
        'port': 389,
        'server': 'localhost',
        'port1': 0,
        'port2': 0,
    },
    'General': {
        # clientport is not needed for Python X2Go
        'clientport': 22, 
        'autoresume': True,
    },
    'Authorization': {
        'newprofile': True,
        'suspend': True,
        'editprofile': True,
        'resume': True
    },
    'trayicon': {
        'enabled': True,
        'mintotray': True,
        'noclose': True,
        'mincon': True,
        'maxdiscon': True,
    },
}
X2GO_CLIENTPRINTING_DEFAULTS = {
    'General': {
        # showdialog will result in a print action that allows opening a print dialog box
        'showdialog': False,
        # if true, open a PDF viewer (or save as PDF file). If false, print via CUPS or print command
        'pdfview': True,
    },
    'print': {
        # If false, print via CUPS. If true, run "command" to process the print job
        'startcmd': False,
        # print command for non-CUPS printing
        'command': 'lpr',
        # ignored in Python X2Go
        'stdin': False,
        # ignored in Python X2Go
        'ps': False,
    },
    'save': {
        # a path relative to the user's home directory
        'folder': 'PDF',
    },
    'view': {
        # If General->pdfview is true: 
        #   if open is true, the PDF viewer command is executed
        #   if open is false, the incoming print job is saved in ~/PDF folder 
        'open': True,
        # command to execute as PDF viewer
        'command': 'xpdf',
    },
    'CUPS': {
        # default print queue for CUPS, if print queue does not exist, the default 
        # CUPS queue is detected
        'defaultprinter': 'PDF',
    },
}
if X2GOCLIENT_OS == 'Windows':
    X2GO_CLIENTPRINTING_DEFAULTS['print'].update({'gsprint': os.path.join(os.environ['ProgramFiles'], 'GhostGum', 'gsview', 'gsprint.exe'), })


if X2GOCLIENT_OS == 'Windows':
    X2GO_CLIENTXCONFIG_DEFAULTS = {
        'XServers': {
            'known_xservers': ['VcXsrv_development', 'VcXsrv_shipped', 'VcXsrv', 'Xming', 'Cygwin-X', ],
        },
        'Cygwin-X': {
            'display': 'localhost:40',
            'last_display': 'localhost:40',
            'process_name': 'XWin.exe',
            'test_installed': os.path.join(os.environ['SystemDrive'], '\\', 'cygwin', 'bin', 'XWin.exe'),
            'run_command': os.path.join(os.environ['SystemDrive'], '\\', 'cygwin', 'bin', 'XWin.exe'),
            'parameters': [':40', '-clipboard', '-multiwindow', '-notrayicon', '-nowinkill', '-nounixkill', '-swcursor', ],
        },
        'VcXsrv': {
            'display': 'localhost:40',
            'last_display': 'localhost:40',
            'process_name': 'vcxsrv.exe',
            'test_installed': os.path.join(os.environ['ProgramFiles'], 'VcXsrv', 'vcxsrv.exe'), 
            'run_command': os.path.join(os.environ['ProgramFiles'], 'VcXsrv', 'vcxsrv.exe'),
            'parameters': [':40', '-clipboard', '-multiwindow', '-notrayicon', '-nowinkill', '-nounixkill', '-swcursor', ],
        },
        'VcXsrv_shipped': {
            'display': 'localhost:40',
            'last_display': 'localhost:40',
            'process_name': 'vcxsrv.exe',
            'test_installed': os.path.join(os.getcwd(), 'VcXsrv', 'vcxsrv.exe'), 
            'run_command': os.path.join(os.getcwd(), 'VcXsrv', 'vcxsrv.exe'),
            'parameters': [':40', '-clipboard', '-multiwindow', '-notrayicon', '-nowinkill', '-nounixkill', '-swcursor', ],
        },
        'VcXsrv_development': {
            'display': 'localhost:40',
            'last_display': 'localhost:40',
            'process_name': 'vcxsrv.exe',
            'test_installed': os.path.join(os.getcwd(), '..', 'pyhoca-contrib', 'mswin', 'vcxsrv-mswin', 'VcXsrv-1.14.2.0', 'vcxsrv.exe'), 
            'run_command': os.path.join(os.getcwd(), '..', 'pyhoca-contrib', 'mswin', 'vcxsrv-mswin', 'VcXsrv-1.14.2.0', 'vcxsrv.exe'), 
            'parameters': [':40', '-clipboard', '-multiwindow', '-notrayicon', '-nowinkill', '-nounixkill', '-swcursor', ],
        },
        'Xming': {
            'display': 'localhost:40',
            'last_display': 'localhost:40',
            'process_name': 'Xming.exe',
            'test_installed': os.path.join(os.environ['ProgramFiles'], 'Xming', 'Xming.exe'), 
            'run_command': os.path.join(os.environ['ProgramFiles'], 'Xming', 'Xming.exe'),
            'parameters': [':40', '-clipboard', '-multiwindow', '-notrayicon', '-nowinkill', '-nounixkill', '-swcursor', ],
        },
    }
else:
    # make the variable available when building API documentation with epydoc
    X2GO_CLIENTXCONFIG_DEFAULTS = {}

X2GO_GENERIC_APPLICATIONS = [ 'WWWBROWSER', 'MAILCLIENT', 'OFFICE', 'TERMINAL', ]
"""X2Go's generic applications."""

X2GO_SESSIONPROFILE_DEFAULTS = {
    'autologin': True, 'autoconnect': False, 'autostart': False, 'setsessiontitle': False, 'sessiontitle': "",
    'speed': 2, 'pack': '16m-jpeg', 'quality': 9,
    'iconvto': 'UTF-8', 'iconvfrom': 'UTF-8', 'useiconv': False,
    'usesshproxy': False, 'sshproxyhost': 'proxyhost.mydomain', 'sshproxyport': 22, 'sshproxyuser': '', 'sshproxykeyfile': '',
    'sshproxytype': 'SSH', 'sshproxysameuser': False, 'sshproxysamepass': False, 'sshproxyautologin': True,
    'uniquehostkeyaliases': False,
    'useexports': True, 'restoreexports': False, 'fstunnel': True, 'export': '',
    'usemimebox': False, 'mimeboxextensions': '', 'mimeboxaction': 'OPEN',
    'fullscreen': False,
    'width': 800,'height': 600, 'maxdim': False, 'dpi': 96, 'setdpi': False, 'xinerama': False, 'multidisp': False, 'display': 1,
    'usekbd': True, 'layout': 'us', 'type': 'pc105/us', 'variant': '',
    'sound': False, 'soundsystem': 'pulse', 'startsoundsystem': False, 'soundtunnel':True, 'defsndport':True, 'sndport':4713,
    'name': 'NEW_PROFILE', 'icon': ':icons/128x128/x2gosession.png',
    'host': 'server.mydomain', 'user': CURRENT_LOCAL_USER, 'key': '', 'sshport': 22, 'krblogin': False, 'forwardsshagent': False,
    'rootless': True, 'applications': X2GO_GENERIC_APPLICATIONS, 'command':'TERMINAL', 'published': False,
    'directrdp': False, 'directrdpsettings': '', 'rdpclient': 'rdesktop', 'rdpport': 3389,
    'rdpoptions': '-u X2GO_USER -p X2GO_PASSWORD', 'rdpserver': '',
    'print': False,
    'xdmcpserver': 'localhost',
}
"""L{X2GoSessionProfiles} default values to fill a new session profile with."""
##
## X2Go Proxy defaults
##

# here is a list of NX 3.x compression methods, this is the "%"-hashed list that
# can also be used for printing in help texts, docs etc.
# The "%"-sign can be replaced by digits 0-9.
pack_methods_nx3_noqual  = ['nopack','8','64','256','512','4k','32k','64k','256k','2m','16m',
                            '256-rdp','256-rdp-compressed','32k-rdp','32k-rdp-compressed','64k-rdp',
                            '64k-rdp-compressed','16m-rdp','16m-rdp-compressed',
                            'rfb-hextile','rfb-tight','rfb-tight-compressed',
                            '8-tight','64-tight','256-tight','512-tight','4k-tight','32k-tight',
                            '64k-tight','256k-tight','2m-tight','16m-tight',
                            '8-jpeg-%','64-jpeg','256-jpeg','512-jpeg','4k-jpeg','32k-jpeg',
                            '64k-jpeg','256k-jpeg','2m-jpeg','16m-jpeg-%',
                            '8-png-jpeg-%','64-png-jpeg','256-png-jpeg','512-png-jpeg','4k-png-jpeg',
                            '32k-png-jpeg','64k-png-jpeg','256k-png-jpeg','2m-png-jpeg','16m-png-jpeg-%',
                            '8-png-%','64-png','256-png','512-png','4k-png',
                            '32k-png','64k-png','256k-png','2m-png','16m-png-%',
                            '16m-rgb-%','16m-rle-%',]
"""Available NX3 compression methods."""

# use for printing on screen...
pack_methods_nx3_formatted="""
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
    \'%s\'
""" % ('\', \''.join(pack_methods_nx3_noqual[0:11]), \
           '\', \''.join(pack_methods_nx3_noqual[11:16]), \
           '\', \''.join(pack_methods_nx3_noqual[16:19]), \
           '\', \''.join(pack_methods_nx3_noqual[19:22]), \
           '\', \''.join(pack_methods_nx3_noqual[22:28]), \
           '\', \''.join(pack_methods_nx3_noqual[28:32]), \
           '\', \''.join(pack_methods_nx3_noqual[32:38]), \
           '\', \''.join(pack_methods_nx3_noqual[38:42]), \
           '\', \''.join(pack_methods_nx3_noqual[42:47]), \
           '\', \''.join(pack_methods_nx3_noqual[47:52]), \
           '\', \''.join(pack_methods_nx3_noqual[52:57]), \
           '\', \''.join(pack_methods_nx3_noqual[57:62]), \
           '\', \''.join(pack_methods_nx3_noqual[62:]))

# pack_methods_nx3 is the complete list of NX3 pack methods that can be used to check options 
# against
pack_methods_nx3 = [ m for m in pack_methods_nx3_noqual if "%" not in m ]
for meth in [ m for m in pack_methods_nx3_noqual if "%" in m ]:
    pack_methods_nx3 += [ meth.replace('%','%s' % str(i)) for i in range(0,10) ]
pack_methods_nx3.sort()
##
## X2Go session defaults
##

X2GO_DESKTOPSESSIONS={
    'CINNAMON': 'cinnamon',
    'KDE': 'startkde',
    'GNOME': 'gnome-session',
    'MATE': 'mate-session',
    'XFCE': 'xfce4-session',
    'LXDE': 'startlxde',
    'TRINITY': 'starttrinity',
    'UNITY': 'unity',
}
"""A dictionary with meta-commands for X2Go's window manager sessions."""

##
## X2Go SFTP server defaults
##

RSAKEY_STRENGTH = 1024
RSAHostKey = paramiko.RSAKey.generate(RSAKEY_STRENGTH)
"""\
An RSA host key for this client session. Python X2Go does not use the
system's host key but generates its own host key for each running
application instance.

"""

X2GO_PRINT_ACTIONS = {
    'PDFVIEW': 'X2GoPrintActionPDFVIEW',
    'PDFSAVE': 'X2GoPrintActionPDFSAVE',
    'PRINT': 'X2GoPrintActionPRINT',
    'PRINTCMD': 'X2GoPrintActionPRINTCMD',
    'DIALOG': 'X2GoPrintActionDIALOG',
}
"""Relating print action names and classes."""

DEFAULT_PDFVIEW_CMD = 'xdg-open'
"""Default PDF viewer command for Linux systems (PDFVIEW print action)."""
DEFAULT_PDFSAVE_LOCATION = 'PDF'
"""Default location for saving PDF files (PDFSAVE print action)."""
DEFAULT_PRINTCMD_CMD = 'lpr'
"""Default command for the PRINTCMD print action."""

X2GO_MIMEBOX_ACTIONS = {
    'OPEN': 'X2GoMIMEboxActionOPEN',
    'OPENWITH': 'X2GoMIMEboxActionOPENWITH',
    'SAVEAS': 'X2GoMIMEboxActionSAVEAS',
}
"""Relating MIME box action names and classes."""

X2GO_MIMEBOX_EXTENSIONS_BLACKLIST = [
    'LOCK', 'SYS', 'SWP', 
    'EXE', 'COM', 'CMD', 'PS1', 'PS2', 'BAT',
    'JS', 'PY', 'PL', 'SH',
]
"""Black-listed MIME box file extenstions."""

# X2Go desktop sharing
X2GO_SHARE_VIEWONLY=0
"""Constant representing read-only access to shared desktops."""
X2GO_SHARE_FULLACCESS=1
"""Constant representing read-write (full) access to shared desktops."""

PUBAPP_MAX_NO_SUBMENUS=10
"""Less than ten applications will not get rendered into submenus."""
